//Ethan Alexander Shulman 2024

var aiWorker, aiDelays, aiTimers, mapStateBuffer, currentMaxAiIndex = 0,
aiStateMsg = {"msg":1, "s":null}, aiTickMsg = {"msg":2, "a":0};

const WMSG_GAME_DATA = 0,
WMSG_UPDATE_STATE = 1,
WMSG_AI_TICK = 2,
WMSG_START_RECORDING = 10,
WMSG_RECORD_GAME = 11,
WMSG_TRAIN_NEURALNET = 12,

MMSG_GAME_LOADED = 0,
MMSG_AI_ACTION = 1,
MMSG_RECORDING = 10,


AI_INFO = [
["Doofus", "wanders around randomly with no clear strategy.", 0],
["Plushkin", "hoards what little Pixels it has, very rarely does it risk capturing new Mines.", 0],
["Mus", "darts erratically from point to point behaving as a horde.", 0],
["Unimodus", "distributes Pixels evenly across its domain, slowly spreading.", 0],

["Voidling", "randomly spreads out.", 1],
["Void Guard", "randomly spreads out to connected Mines.", 1],
["Striker", "launches out in a lance striking specific points.", 1],
["Globuulus", "spreads out and accumulates Pixels in various spots forming globs.", 1],

["Adad", "sends flurries of Pixels to capture Mines.", 2],
["Void Lord", "randomly targets Mines and sends large forces to capture them.", 2],
["Leviathan", "amasses Pixels into a ball, encompassing the map as it spreads.", 2],
["Cell", "expands outward building a wall around mines it absorbs.", 2],

["Ra", "expands outward in rays targetting a whole area of Mines.", 3],
["Achilles", "targets mines based off path of least resistance and launches an optimized assault.", 3],
["Kali", "efficiently spreads death across the battlefield.", 3],
["Chronos", "compares possible futures and chooses the actions leading to the best one.", 3]
],

AI_NAMES = new Array(AI_INFO.length);

for (var i = 0; i < AI_INFO.length; i++) AI_NAMES[i] = AI_INFO[i][0];



function InitializeAI() {
	//init ai web worker
	var aiSrc = new Blob([DGE("AIWorkerCode").textContent], {"type":"text/plain"}),
		aiURL = URL.createObjectURL(aiSrc);
	aiWorker = new Worker(aiURL);
	URL.revokeObjectURL(aiURL);
	
	//on ai worker message
	aiWorker.onmessage = function(e) {
		switch (e.data["msg"]) {
			case MMSG_GAME_LOADED:
				//game finished loading
				FinishLaunchingGame();
				break;
				
			case MMSG_AI_ACTION:
				//apply ai action
				AIMovePixels(e.data);
				break;
				
			
			case MMSG_RECORDING:
				if (DEV_VERSION) RecordGame(e.data);
				break;
		}
	};
}


//load game data into ai worker, 'ais' format [[aiIndex, aiDelay, aiSeed]]
function LoadAI(ais) {
	//active factions
	var facs = new Array(activeFactions.length);
	for (var i = 0; i < facs.length; i++) facs[i] = activeFactions[i].id;

	//map points
	var pts = new Array(mapPoints.length);
	mapStateBuffer = new Array(mapPoints.length);
	aiStateMsg["s"] = mapStateBuffer;
	for (var i = 0; i < pts.length; i++) {
		var p = mapPoints[i],
			cons = new Array(p.connections.length);
		for (var k = 0; k < cons.length; k++) {
			cons[k] = p.connections[k].pointId;
		}
		pts[i] = {"x":p.x, "y":p.y, "i":p.pointId, "c":cons, "m":p.mineSize, "f":p.findPath, "l":p.pfLengths};
		mapStateBuffer[i] = new Array(FACTION_COUNT*2);
	}
	
	//spawn points
	var spawns = new Array(activeFactions.length);
	for (var i = 0; i < spawns.length; i++) spawns[i] = mapFactionSpawns[i].pointId;
	
	//ai delays
	aiDelays = new Array(ais.length);
	currentMaxAiIndex = 0;
	for (var i = 0; i < ais.length; i++) {
		aiDelays[i] = Math.ceil(ais[i][1]*TICK_RATE);
		currentMaxAiIndex = Math.max(currentMaxAiIndex, ais[i][0]);
	}
	
	aiTimers = new Array(ais.length);
	aiTimers.fill(0);
	
	aiWorker.postMessage({"msg":WMSG_GAME_DATA, "facs":facs, "pts":pts, "spawns":spawns, "ais":ais});
}


//apply ai move pixels action
function AIMovePixels(data) {
	if (aiRecording && appStatus !== APPSTATUS_GAME) return;

	var aiId = data["i"],
		src = mapPoints[data["a"]],
		dst = mapPoints[data["b"]],
		moveCount = data["c"],
		aiFac = activeFactions[aiId].id,
		mapX = Math.floor(src.x*PIXELMAP_CELL_SCALE),
		mapY = Math.floor(src.y*PIXELMAP_CELL_SCALE),
		searchExt = 0;

	//select pixels for ai movement by searching cells starting from center of point
	while (true) {
		var foundOne = false,
			aRange = (searchExt ? 4 : 1);
		for (var a = 0; a < aRange; a++) {
			var yDir = Math.floor(a/2),
				sign = (a%2)*2-1, minExt = -searchExt, maxExt = searchExt+1,
				gX = 0, gY = 0;
			if (yDir) {
				minExt++;
				maxExt--;
			}
			for (var i = minExt; i < maxExt; i++) {
				if (yDir) {
					gX = mapX+searchExt*sign;
					gY = mapY+i;
				} else {
					gX = mapX+i;
					gY = mapY+searchExt*sign;
				}
				if (gX < 0 || gY < 0 ||
					gX >= PIXELMAP_CELLS || gY >= PIXELMAP_CELLS) continue;
					
				var cell = pixelMap[gX+gY*PIXELMAP_CELLS],
					cellLen = cell.length;
				for (var c = 0; c < cellLen; c++) {
					var cp = cell[c];
					if (cp.faction === aiFac && cp.orbitPoint === src) {
						foundOne = true;
						abuf.push(cp);
						if (abuf.length >= moveCount) {
							i = maxExt;
							a = 4;
							break;
						}
					}
				}
			}
		}
		if (!foundOne || abuf.length >= moveCount) {
			break;
		}
		searchExt++;
	}
	
	//move selected pixels
	PixelsTravelTarget(abuf, dst);
	abuf.length = 0;
}


//update ai sending tick messages
function UpdateAI() {
	var initMap = true;
	for (var i = 0; i < aiTimers.length; i++) {
		if ((++aiTimers[i]) >= aiDelays[i]) {
			aiTimers[i] = 0;
			
			if (initMap) {
				//load faction counts into map state buffer
				var nPts = mapPoints.length;
				for (var n = 0; n < nPts; n++) {
					var p = mapPoints[n], f = p.factionCount, t = p.travelCount,
						b = mapStateBuffer[n], bi = 0;
					for (var k = 0; k < FACTION_COUNT; k++) {
						b[bi++] = f[k];
					}
					for (var k = 0; k < FACTION_COUNT; k++) {
						b[bi++] = t[k];
					}
				}
				aiWorker.postMessage(aiStateMsg);
				initMap = false;
			}
			
			//send ai tick message
			aiTickMsg["a"] = i;
			aiWorker.postMessage(aiTickMsg);
		}
	}
}


//ai recording for training neuralnets
var aiRecording = false,
aiRecordDelay = 0,
aiRecordIds = null,
aiRecordMask = null,
aiRecordFactionCount = 0,
aiRecordGameCount = 0,
aiRecordGameLimit = 0,
aiRecordTimeLimit = 0,
aiRecordTimer = 0,
aiRecordWinner = 0,
aiRecordWinThreshold = 0,
aiRecordAvgPixels = 0,
aiRecordPtr = 0, aiRecordIB = null, aiRecordFB = null,
aiRecordWinLogging = false,
aiRegisteredUnloadWarning = false;


//auto play games recording ai moves for training neuralnets
function RecordAI(delay, aiIds, recordMask, timeLimit, gameLimit, winThreshold) {
	if (!DEV_VERSION) return;

	aiRecordDelay = delay;
	aiRecordIds = aiIds;
	aiRecordMask = recordMask;
	aiRecordWinThreshold = (winThreshold ? winThreshold : 0);
	aiRecordFactionCount = 4;
	aiRecordGameLimit = gameLimit;
	aiRecordTimeLimit = timeLimit;
	aiRecordPtr = aiRecordAvgPixels = 0;
	aiRecordIB = new Uint32Array(1024);
	aiRecordFB = new Float32Array(aiRecordIB.buffer);
	aiRecording = true;
	
	aiWorker.postMessage({"msg":WMSG_START_RECORDING, "mask":recordMask});
	
	EV(mainMenuPanel, false);
	AIRecordingLaunchNextGame();
	AIWindowCloseWarning();
}

function AIRecordingCheckGameOver() {
	if (!DEV_VERSION) return;

	var facLen = activeFactions.length,
		maxAlive = 0, maxI = -1, 
		maxAlive2 = 0;
	for (var i = 0; i < facLen; i++) {
		var nPx = activeFactions[i].pixels.length;
		if (aiRecordMask[i]) {
			if (nPx > maxAlive) {
				maxAlive2 = maxAlive;
				maxAlive = nPx;
				maxI = i;
				
			} else if (nPx > maxAlive2) {
				maxAlive2 = nPx;
			}
		}
	}
	
	if (++aiRecordTimer >= aiRecordTimeLimit || maxAlive2 === 0) {
		if (aiRecordWinLogging) console.log("Winner "+maxI+"/"+activeFactions.length+" with "+maxAlive+", next best "+maxAlive2+" = "+(maxAlive/Math.max(1,maxAlive2)));
		
		var winner = (maxAlive >= aiRecordWinThreshold ? maxI : -1);
		aiRecordTimer = 0;
		appStatus = APPSTATUS_MAINMENU;
		if (winner === -1) {
			AIRecordingLaunchNextGame();
		
		} else {
			aiRecordAvgPixels += maxAlive;
			aiRecordWinner = winner;
			aiWorker.postMessage({"msg":WMSG_RECORD_GAME, "winner":winner});
		}
	}
}


function AIRecordingLaunchNextGame() {
	ClearMap();
	ClearPixels();
	
	aiRecordFactionCount = (aiRecordFactionCount+1)%5;

	var facCount = 2+aiRecordFactionCount,
		ais = [];
	for (var i = 0; i < facCount; i++) {
		ais.push([aiRecordIds[i], aiRecordDelay, Math.round(Math.random()*0xFFFFFFFF)]);
	}
	LaunchGame(2, Math.round(Math.random()*0xFFFFFFFF), ais, 0);
}

function RecordGame(dat) {
	if (!DEV_VERSION) return;

	var moves = dat["m"], states = dat["s"];
	if (!moves || !states) {
		AIRecordingLaunchNextGame();
		return;
	}

	//record data into typed array to conserve memory
	var nPts = mapPoints.length,
		reqSz = aiRecordPtr + 4 + nPts*(3+nPts) + moves.length/4*(5+nPts*3);
		
	//expand binary buffer if needed
	if (reqSz > aiRecordIB.length) {
		var nSz = aiRecordIB.length;
		while (nSz < reqSz) nSz *= 2;
		
		var nIB = new Uint32Array(nSz);
		nIB.set(aiRecordIB);
		aiRecordIB = nIB;
		aiRecordFB = new Float32Array(nIB.buffer);
	}
	
	//record data as binary
	var rp = aiRecordPtr;
	rp++;
	aiRecordIB[rp++] = nPts;
	aiRecordIB[rp++] = moves.length/4;
	
	var moveLen = moves.length,
		afLen = activeFactions.length,
		stateStride = nPts*afLen*2;
	for (var i = 0; i < moveLen; i++) {
		var tick = moves[i++];
		aiRecordIB[rp++] = moves[i++];
		aiRecordIB[rp++] = moves[i++];
		aiRecordFB[rp++] = moves[i];
		
		var normalPtr = rp,
			sti = tick*stateStride,
			maxPixelCount = 0,
			maxSurplus = 0;
		rp += 2;
		for (var sp = 0; sp < nPts; sp++) {
			var nfriend = 0, tfriend = 0, nenemy = 0;
			for (var f = 0; f < afLen; f++) {
				var fc = states[sti++],
					tc = states[sti++];
				maxPixelCount = Math.max(maxPixelCount, fc+tc);
				if (f === aiRecordWinner) {
					nfriend = fc;
					tfriend = tc;
				} else {
					nenemy += fc+tc;
				}
			}
			aiRecordIB[rp++] = nfriend;
			aiRecordIB[rp++] = tfriend;
			aiRecordIB[rp++] = nenemy;
			
			maxSurplus = Math.max(maxSurplus, nfriend-mapPoints[sp].mineSize);
		}
		aiRecordFB[normalPtr++] = (maxPixelCount ? 1/maxPixelCount : 0);
		aiRecordFB[normalPtr] = (maxSurplus ? 1/maxSurplus : 0);
	}
	
	for (var i = 0; i < nPts; i++) {
		var p = mapPoints[i];
		aiRecordFB[rp++] = p.x;
		aiRecordFB[rp++] = p.y;
		aiRecordIB[rp++] = p.mineSize;		
		for (var k = 0; k < nPts; k++) aiRecordFB[rp++] = p.pfLengths[k];
	}
	
	aiRecordIB[aiRecordPtr] = rp-aiRecordPtr;
	aiRecordPtr = rp;
	
	console.log("Recorded game #"+(++aiRecordGameCount)+", total size "+(aiRecordPtr*4/1000/1000)+"MB.");
	
	if (aiRecordGameCount >= aiRecordGameLimit) {
		//done show recording download
		aiRecordIB[aiRecordPtr++] = 0;
		var blob = new Blob([aiRecordIB.subarray(0,aiRecordPtr)]),
			dload = document.createElement("a");
		dload.style.cssText = "font-size: 10vw; position: absolute; left: 0px; top: 0px; z-index: 100;";
		dload.download = "AIRecording.bin";
		dload.textContent = "Download Recording";
		dload.href = URL.createObjectURL(blob);
		document.body.appendChild(dload);
		console.log("Recording complete! Click download to download recording.\n"+
					"Average winner pixels "+Math.round(aiRecordAvgPixels/aiRecordGameLimit)+".");
		aiRecording = false;
		
	} else {
		AIRecordingLaunchNextGame();
	}
}

//preemptively stop ai recording and prompt download
function AIForceStopAndDownload() {
	if (!DEV_VERSION) return;

	aiRecordTimer = 0;
	appStatus = APPSTATUS_MAINMENU;
	
	aiRecordIB[aiRecordPtr++] = 0;
	var blob = new Blob([aiRecordIB.subarray(0,aiRecordPtr)]),
		dload = document.createElement("a");
	dload.style.cssText = "font-size: 10vw; position: absolute; left: 0px; top: 0px; z-index: 100;";
	dload.download = "AIRecording.bin";
	dload.textContent = "Download Recording";
	dload.href = URL.createObjectURL(blob);
	document.body.appendChild(dload);
	console.log("Recording complete! Click download to download recording.\n"+
				"Average winner pixels "+Math.round(aiRecordAvgPixels/aiRecordGameCount)+" over "+aiRecordGameCount+" games.");
	aiRecording = false;
}


//launch neural network training
function LaunchNeuralNetworkTraining(srcHiddenLayers,srcWeightScale,srcLearnRate,srcRetryThreshold,srcErrorNoise,
dstHiddenLayers,dstWeightScale,dstLearnRate,dstRetryThreshold,dstErrorNoise, iterations) {
	if (!DEV_VERSION) return;

	var file = document.createElement("input");
	file.type = "file";
	file.multiple = true;
	file.onchange = function(e) {
		var filei = e.target;
		if (filei.files && filei.files.length) {
			var loaded = [],
			onloadFunc = function(re) {
				loaded.push(re.target.result);
				if (loaded.length >= filei.files.length) {
					console.log("Loaded "+loaded.length+" recordings, initializing.");
					if (loaded.length > 1) {
						var tsz = 1;
						for (var i = 0; i < loaded.length; i++) {
							loaded[i] = new Uint32Array(loaded[i]);
							tsz += loaded[i].length-1;
						}
						var tbuf = new Uint32Array(tsz), offset = 0;
						for (var i = 0; i < loaded.length; i++) {
							tbuf.set(loaded[i].subarray(0,loaded[i].length-1), offset);
							offset += loaded[i].length-1;
						}
						tbuf[offset] = 0;
						loaded[0] = tbuf.buffer;
					}

					aiWorker.postMessage({"msg":WMSG_TRAIN_NEURALNET, "buf":loaded[0],
						"sh":srcHiddenLayers, "sw":srcWeightScale, "sr":srcLearnRate, "st":srcRetryThreshold, "sn":srcErrorNoise,
						"dh":dstHiddenLayers, "dw":dstWeightScale, "dr":dstLearnRate, "dt":dstRetryThreshold, "dn":dstErrorNoise, "iter":iterations});
					
					renderingEnabled = false;
					AIWindowCloseWarning();
				}
			};
			
			for (var fi = 0; fi < filei.files.length; fi++) {
				var lfile = filei.files[fi];
				console.log("Loading "+(lfile.size/1000000)+"MB recording.");
				 
				var fread = new FileReader();
				fread.onload = onloadFunc;
				fread.readAsArrayBuffer(lfile);
			}
		}
	};
	file.click();
}


//warn before exiting page
function AIWindowCloseWarning() {
	if (!DEV_VERSION) return;
	if (aiRegisteredUnloadWarning) return;
	
	window.addEventListener("beforeunload", function(e) {
		var str = "Warning: AI training process may still be running!";
		console.error(str);
		e.preventDefault();
		return str;
	});
	aiRegisteredUnloadWarning = true;
}


//which ai is currently winning
function AIWinning() {
	var facLen = activeFactions.length,
		maxAlive = 0, maxI = -1, 
		maxAlive2 = 0;
	for (var i = 0; i < facLen; i++) {
		var nPx = activeFactions[i].pixels.length;
		if (aiRecordMask[i]) {
			if (nPx > maxAlive) {
				maxAlive2 = maxAlive;
				maxAlive = nPx;
				maxI = i;
				
			} else if (nPx > maxAlive2) {
				maxAlive2 = nPx;
			}
		}
	}
	return maxI+" is winning by "+(maxAlive/Math.max(1,maxAlive2)).toFixed(3)+"x.";
}

//number of add+multiplies to run neuralnet
function NeuralNetworkCost(srcHidden, dstHidden) {
	var inp = 7,
		cost = 0;
	for (var i = 0; i < srcHidden.length; i++) {
		var out = srcHidden[i];
		cost += (inp+1)*out;
		inp = out;
	}
	cost += 1+inp;
	
	inp = 54;
	for (var i = 0; i < dstHidden.length; i++) {
		var out = dstHidden[i];
		cost += (inp+1)*out;
		inp = out;
	}
	cost += 2+2*inp;
	
	return cost*144;
}

