//Ethan Alexander Shulman 2024

var ubuf = new Uint32Array(8), abuf = [], bbuf = [],
activeFactions, spawnPoints, mapPoints = [], mapMines = [], ais = [],
aiAction = {"msg":1,"i":0,"a":0,"b":0,"c":0},
aiClasses;

const MAX_FACTIONS = 6,

WMSG_GAME_DATA = 0,
WMSG_UPDATE_STATE = 1,
WMSG_AI_TICK = 2,
WMSG_START_RECORDING = 10,
WMSG_RECORD_GAME = 11,
WMSG_TRAIN_NEURALNET = 12,

MMSG_GAME_LOADED = 0,
MMSG_AI_ACTION = 1,
MMSG_RECORDING = 10;


//declare ai classes order
function DeclareAIClasses() {
	aiClasses = [DoofusNN, PlushkinNN, MusNN, Unimodus,
				VoidlingNN, VoidGuard, StrikerNN, GlobuulusNN,
				AdadNN, VoidLord, Leviathan, CellAI,
				RaNN, Achilles, KaliNN, Chronos];
}


//uint32 prng hash
function PRNGHash() {
	ubuf[0] ^= Math.imul(ubuf[0],2425714368) ^ (ubuf[0]>>>1);
	ubuf[0] ^= Math.imul(ubuf[0],2893397823) ^ (ubuf[0]>>>10);
	ubuf[0] ^= Math.imul(ubuf[0],2058600894) ^ (ubuf[0]>>>19);
}


/**@constructor*/
function MapPoint(x,y,pi,cons,mine,fpath,plens) {
	this.x = x;
	this.y = y;
	this.pointId = pi;
	this.connections = cons;
	this.mineSize = mine;
	this.mineFaction = -1;
	this.factionCount = new Array(MAX_FACTIONS);
	this.travelCount = new Array(MAX_FACTIONS);
	this.findPath = fpath;
	this.pointLengths = plens;
	this.maxPointLength = 0;
	
	for (var i = 0, l = plens.length; i < l; i++) this.maxPointLength = Math.max(this.maxPointLength,plens[i]);
}


/**@constructor*/
function AI(pId,seed) {
	this.playerId = pId;
	this.faction = activeFactions[pId];
	this.spawn = spawnPoints[pId];
	this.seed = seed;
}

AI.prototype.Decide = function() {}

//ai move pixels command
AI.prototype.MovePixels = function(srcPt,dstPt,count) {
	aiAction["i"] = this.playerId;
	aiAction["a"] = srcPt.pointId;
	aiAction["b"] = dstPt.pointId;
	aiAction["c"] = count;
	self.postMessage(aiAction);
	
	if (DEV_VERSION && aiRecording) aiRecording.RecordMove(this.playerId,srcPt,dstPt,count);
}

//random int between min and max(exclusive)
AI.prototype.Rand = function(minv,maxv) {
	ubuf[0] = this.seed++;
	if (this.seed > 0xFFFFFFFF) this.seed = 0;
	PRNGHash();
	return ubuf[0]%(maxv-minv) + minv;
}

//random float between min and max(inclusive)
AI.prototype.RandFloat = function(minv,maxv) {
	ubuf[0] = this.seed++;
	if (this.seed > 0xFFFFFFFF) this.seed = 0;
	PRNGHash();
	return (ubuf[0]/0xFFFFFFFF)*(maxv-minv)+minv;
}



//load game data
function LoadGame(data) {
	//active factions
	activeFactions = data["facs"];

	//map points
	var pts = data["pts"], len = pts.length;
	mapPoints.length = 0;
	for (var i = 0; i < len; i++) {
		var p = pts[i];
		mapPoints.push(new MapPoint(p["x"],p["y"],p["i"],p["c"],p["m"],p["f"],p["l"]));
	}
	
	//mines
	mapMines.length = 0;
	for (var i = 0; i < len; i++) {
		var p = mapPoints[i];
		if (p.mineSize > 0) mapMines.push(p);
	}
	
	//spawn points
	var spawns = data["spawns"];
	spawnPoints = spawns;
	for (var i = 0; i < spawns.length; i++) {
		spawns[i] = mapPoints[spawns[i]];
	}
	
	//ais
	var aiDat = data["ais"];
	ais.length = 0;
	nnaiPresent = false;
	for (var i = 0; i < aiDat.length; i++) {
		var ad = aiDat[i],
			aid = i+1;
		if (DEV_VERSION && aiRecording) aid = i;
		ais.push(new aiClasses[ad[0]](aid,ad[2]));
	}
	
	if (DEV_VERSION && aiRecording) aiRecording.Clear();
	
	//confirm finished
	self.postMessage({"msg": MMSG_GAME_LOADED});
}


//update map faction counts
function UpdateState(data) {
	var state = data["s"],
		nPts = state.length;
	for (var i = 0; i < nPts; i++) {
		var b = state[i], bi = 0,
			p = mapPoints[i], a = p.factionCount;
		for (var k = 0; k < MAX_FACTIONS; k++) a[k] = b[bi++];
		a = p.travelCount;
		for (var k = 0; k < MAX_FACTIONS; k++) a[k] = b[bi++];
	}
	
	nnaiViewOld = true;
	if (DEV_VERSION && aiRecording) aiRecording.RecordState();
}


//on receive worker message
self.onmessage = function(e) {
	switch (e.data["msg"]) {
		case WMSG_GAME_DATA:
			LoadGame(e.data);
			break;
			
		case WMSG_UPDATE_STATE:
			UpdateState(e.data);
			break;
			
		case WMSG_AI_TICK:
			ais[e.data["a"]].Decide();
			break;
			
			
		case WMSG_START_RECORDING:
			if (DEV_VERSION) StartRecording(e.data);
			break;
			
		case WMSG_RECORD_GAME:
			if (DEV_VERSION) aiRecording.RecordGame(e.data);
			break;
			
		case WMSG_TRAIN_NEURALNET:
			if (DEV_VERSION) TrainNeuralNet(e.data);
			break;
	}
};


var plsLengths;
function PointLengthSort(a,b) {
	return plsLengths[a.pointId] > plsLengths[b.pointId] ? 1 : -1;
}


/**@constructor
@extends {AI}*/
function Unimodus(pId,seed) {
	AI.call(this,pId,seed);

	this.domainOrder = mapPoints.slice();
	plsLengths = this.spawn.pointLengths;
	this.domainOrder.sort(PointLengthSort);

	this.domainId = 0;
	this.domainRange = 0;
}
Unimodus.prototype = Object.create(AI.prototype);

Unimodus.prototype.Decide = function() {
	//find biggest difference between neighbour pixels
	var domOrder = this.domainOrder, domLen = this.domainId+1,
		domRange = this.domainRange,
		nPts = mapPoints.length,
		pfLens = this.spawn.pointLengths,
		facId = this.faction,
		maxDiff = 0, maxA, maxB,
		minCount = 1e7, maxMineSz = 0, allFull = true;
	for (var i = 0; i < domLen; i++) {
		var pt = domOrder[i],
			fcount = pt.factionCount[facId],
			tcount = pt.travelCount[facId],
			count = fcount+tcount,
			cons = pt.connections,
			mineSz = pt.mineSize,
			maxTransfer = fcount-mineSz;
		minCount = Math.min(minCount, count);
		if (mineSz > 0) {
			maxMineSz = Math.max(maxMineSz, mineSz);
			if (fcount < mineSz) allFull = false;
		}
		
		for (var k = 0; k < cons.length; k++) {
			var ci = cons[k];
			if (pfLens[ci] <= domRange) {
				var cpt = mapPoints[ci],
					ccount = cpt.factionCount[facId]+cpt.travelCount[facId],
					diff = (count-ccount)*0.5;
				if (mineSz > 0) {
					diff = Math.min(diff, maxTransfer);
				} else {
					if (diff >= 0.5) diff += this.Rand(0,2)*0.5;
				}
				diff = Math.floor(diff);
				if (diff > maxDiff) {
					maxDiff = diff;
					maxA = pt;
					maxB = cpt;
				}
			}
		}
	}
	
	if (allFull && minCount >= maxMineSz) {
		//if all mines in domain full, extend domain
		if (domLen < nPts) {
			this.domainId = domLen;
			this.domainRange = pfLens[domOrder[domLen].pointId];
		}
	}
	
	if (maxDiff > 0) {
		//spread pixels uniformly across domain, by moving pixels at the largest difference we found
		this.MovePixels(maxA,maxB,maxDiff);
	}
}



/**@constructor
@extends {AI}*/
function VoidGuard(pId,seed) {
	AI.call(this,pId,seed);
}
VoidGuard.prototype = Object.create(AI.prototype);

VoidGuard.prototype.Decide = function() {
	//find largest 'surplus' to choose movement from
	var nPts = mapPoints.length,
		facId = this.faction,
		maxSurplus = 0, maxPt = null,
		i;
	for (i = 0; i < nPts; i++) {
		var pt = mapPoints[i],
			count = pt.factionCount[facId]-pt.mineSize;
		if (count > maxSurplus) {
			maxSurplus = count;
			maxPt = pt;
		}
	}
	
	maxSurplus = Math.round(maxSurplus*0.5);
	if (maxSurplus > 1) {
		//assign each possible movement(neighbour) a probability
		var cons = maxPt.connections,
			optLen = cons.length,
			total = 0;
		for (i = 0; i < optLen; i++) {
			var cpt = mapPoints[cons[i]],
				fcount = cpt.factionCount[facId]+cpt.travelCount[facId],
				prob = (fcount < cpt.mineSize ? 20 : 1/(1+fcount));
			abuf.push(prob);
			total += prob;
		}
		if (total === 0) {
			abuf.length = 0;
			return;
		}
		
		//randomly select movement option from probabilities
		var rind = this.RandFloat(0,total-1e-6),
			t = 0;
		for (i = 0; i < optLen; i++) {
			var nxt = t+abuf[i];
			if (rind < nxt) break;
			t = nxt;
		}
		abuf.length = 0;
		this.MovePixels(maxPt,mapPoints[cons[i]], maxSurplus);
	}
}



/**@constructor
@extends {AI}*/
function VoidLord(pId,seed) {
	AI.call(this,pId,seed);
	this.nextTarget = null;
}
VoidLord.prototype = Object.create(AI.prototype);


VoidLord.prototype.Decide = function() {
	//find largest 'surplus' to move from
	var nPts = mapMines.length,
		facId = this.faction,
		maxSurplus = 0, maxPt = null,
		i;
	for (i = 0; i < nPts; i++) {
		var pt = mapMines[i],
			count = pt.factionCount[facId]-pt.mineSize;
		if (count > maxSurplus) {
			maxSurplus = count;
			maxPt = pt;
		}
	}
	
	var ntarg = this.nextTarget;
	if (ntarg && ntarg.factionCount[facId] >= ntarg.mineSize) {
		this.nextTarget = ntarg = null;//target full, find new target
	}
	if (!ntarg && maxSurplus > 4) {
		//find new target
		var plens = maxPt.pointLengths,
			total = 0,
			maxPlen = maxPt.maxPointLength;
		for (i = 0; i < nPts; i++) {
			var pt = mapMines[i],
				count = pt.factionCount[facId],
				mineSz = pt.mineSize,
				prob = 0;
			if (count < mineSz) {
				prob = Math.exp((maxPlen-plens[pt.pointId])*0.01);
				total += prob;
			}
			abuf.push(prob);
		}
		
		//randomly select target from probabilities
		if (total > 0) {
			var rind = this.RandFloat(0,total-1e-6),
				t = 0;
			for (i = 0; i < nPts; i++) {
				var nxt = t+abuf[i];
				if (rind < nxt) break;
				t = nxt;
			}
			this.nextTarget = ntarg = mapMines[i];
		}
		abuf.length = 0;
	}
	
	//if we can fill target, send pixels
	if (ntarg && maxSurplus >= ntarg.mineSize) {
		this.MovePixels(maxPt,ntarg, maxSurplus);
		this.nextTarget = null;
	}
}



/**@constructor
@extends {AI}*/
function Leviathan(pId,seed) {
	AI.call(this,pId,seed);
	this.blob = this.spawn;
	this.blobLastCount = 0;
	this.target = null;
	this.debt = 0;
	this.sendTo = 0;
}
Leviathan.prototype = Object.create(AI.prototype);

Leviathan.prototype.Decide = function() {
	var nMines = mapMines.length,
		blobPt = this.blob,
		facId = this.faction,
		maxSurplus = 0, maxSurpI = 0,
		maxSz = 0, maxSzI = 0;
	for (var i = 0; i < nMines; i++) {
		var pt = mapMines[i];
		if (pt === blobPt) continue;
		
		var msz = pt.mineSize,
			surplus = pt.factionCount[facId]-msz;
		if (surplus >= 0) {
			if (surplus > maxSurplus) {
				maxSurplus = surplus;
				maxSurpI = i;
			}
			if (msz > maxSz) {
				maxSz = msz;
				maxSzI = i;
			}
		}
	}
	
	var blobCount = blobPt.factionCount[facId];
	if (blobCount < blobPt.mineSize) {
		//if blob empty, select new blob at largest filled mine
		this.blob = blobPt = mapMines[maxSzI];
		blobCount = blobPt.factionCount[facId];
		
	} else {
		this.debt = Math.min(0, this.debt+Math.max(0, blobCount-this.blobLastCount));
	}
	this.blobLastCount = blobCount;
	var blobSurplus = blobCount-blobPt.mineSize;
	
	//check if target full
	var ntarg = this.target;
	if (ntarg && ntarg.factionCount[facId] >= ntarg.mineSize) {
		if (ntarg.mineSize > blobPt.mineSize) {
			//new biggest use as blob
			this.blob = blobPt = ntarg;
			this.blobLastCount = blobPt.factionCount[facId];
			blobSurplus = 0;
		}
		this.target = ntarg = null;
	}
	if (!ntarg) {
		//find new target
		var plens = blobPt.pointLengths,
			minPlen = 1e8,
			maxPlen = -1;
		for (var i = 0; i < nMines; i++) {
			var pt = mapMines[i],
				count = pt.factionCount[facId],
				mineSz = pt.mineSize;
			if (count < mineSz) {
				var len = plens[pt.pointId];
				minPlen = Math.min(minPlen,len);
				maxPlen = Math.max(maxPlen,len);
			}
		}
		if (maxPlen === -1) return;
			
		var total = 0;
		maxPlen = 1/(maxPlen-minPlen);
		for (var i = 0; i < nMines; i++) {
			var pt = mapMines[i],
				count = pt.factionCount[facId],
				mineSz = pt.mineSize,
				prob = 0;
			if (count < mineSz) {
				prob = mineSz*Math.exp((plens[pt.pointId]-minPlen)*maxPlen*-40+10);
				total += prob;
			}
			abuf.push(prob);
		}
		
		//randomly select target from probabilities
		if (total > 0) {
			var rind = this.RandFloat(0,total-1e-6),
				t = 0, i = 0;
			for (i = 0; i < nMines; i++) {
				var nxt = t+abuf[i];
				if (rind < nxt) break;
				t = nxt;
			}
			this.target = ntarg = mapMines[i];
		}
		abuf.length = 0;
	}
	
	var srcPt = null, dstPt = null, moveSz = 0;
	if (ntarg && this.debt === 0 && Math.max(blobSurplus,maxSurplus) >= ntarg.mineSize) {
		dstPt = ntarg;
		moveSz = ntarg.mineSize;
		if (maxSurplus < moveSz) {
			srcPt = blobPt;
			this.debt -= moveSz;
			
		} else {
			srcPt = mapMines[maxSurpI];
		}
		
	} else if (maxSurplus > 0) {
		srcPt = mapMines[maxSurpI];
		dstPt = blobPt;
		moveSz = maxSurplus;
	}
		
	if (dstPt) {
		this.MovePixels(srcPt,dstPt,moveSz);
	}
}



/**@constructor
@extends {AI}*/
function Achilles(pId,seed) {
	AI.call(this,pId,seed);
	this.nextTarget = null;
	this.scanRange = 0;
}
Achilles.prototype = Object.create(AI.prototype);

Achilles.prototype.Decide = function() {
	//find mine with largest surplus
	var nPts = mapMines.length,
		facId = this.faction,
		maxSurplus = 0, maxPt = null,
		i;
	for (i = 0; i < nPts; i++) {
		var pt = mapMines[i],
			count = pt.factionCount[facId]-pt.mineSize;
		if (count > maxSurplus) {
			maxSurplus = count;
			maxPt = pt;
		}
	}
	
	var ntarg = this.nextTarget;
	if (ntarg && ntarg.factionCount[facId] >= ntarg.mineSize) {
		this.nextTarget = ntarg = null;//target full, find new target
	}
	
	if (!ntarg && maxSurplus > 4) {
		//find mines within scan range, gradually increase scan range until we find new target
		var pathLens = maxPt.pointLengths, maxLen = maxPt.maxPointLength,
			scanRange = this.scanRange+250;
		for (i = 0; i < nPts; i++) {
			var pt = mapMines[i];
			if (pathLens[pt.pointId] < scanRange && pt.factionCount[facId] < pt.mineSize) {
				abuf.push(pt);
			}
		}
		
		if (abuf.length === 0) {
			this.scanRange = scanRange;
			return;
		}
		this.scanRange = 0;
		
		//calculate probabilities of each target based on distance, mine size and enemies on route(resistance)
		var total = 0, 
			alen = abuf.length;
		for (i = 0; i < alen; i++) {
			var pt = abuf[i],
				count = pt.factionCount[facId],
				prob = 0,
				ec = 0, sPt = maxPt, targId = pt.pointId;
			//count enemies on route
			while (sPt !== pt) {
				var np = mapPoints[sPt.connections[sPt.findPath[targId]]],
					nFac = np.factionCount, nTrav = np.travelCount;
				for (var k = 0; k < MAX_FACTIONS; k++) {
					if (k === facId) continue;
					ec += nFac[k]+nTrav[k];
				}
				sPt = np;
			}
			prob = (maxLen-pathLens[targId])*pt.mineSize/(1+ec);
			total += prob;
			bbuf.push(prob);
		}
	
		//choose route randomly from probabilities
		if (total > 0) {
			var rind = this.RandFloat(0,total-1e-6),
				t = 0;
			for (i = 0; i < alen; i++) {
				var nxt = t+bbuf[i];
				if (rind < nxt) break;
				t = nxt;
			}
			this.nextTarget = ntarg = abuf[i];
		}
		abuf.length = bbuf.length = 0;
	}
	
	//send pixels to target
	if (ntarg && maxSurplus > 0) {
		this.MovePixels(maxPt,ntarg, Math.min(ntarg.mineSize,maxSurplus));
	}
}



/**@constructor
@extends {AI}*/
function CellAI(pId,seed) {
	AI.call(this,pId,seed);

	this.mineOrder = mapMines.slice();
	plsLengths = this.spawn.pointLengths;
	this.mineOrder.sort(PointLengthSort);
	
	this.growing = true;
	this.lastOrderIndex = 0;
	this.extension = 0;
}
CellAI.prototype = Object.create(AI.prototype);

CellAI.prototype.Decide = function() {
	var mineOrder = this.mineOrder,
		facId = this.faction;
	if (this.growing) {
		//initially grow cell quickly without wall sending waves of pixels straight between first 3 mines
		var nearEmpty = null, nearDst = 1e6, numEmpty = 0,
			maxPt = null, maxSurplus = 0,
			pfLens = this.spawn.pointLengths;
		for (var i = 0; i < 3; i++) {
			var pt = mineOrder[i],
				count = pt.factionCount[facId]-pt.mineSize;
			if (count < 0) {
				var dst = pfLens[pt.pointId];
				if (dst < nearDst) {
					nearDst = dst;
					nearEmpty = pt;
				}
				numEmpty++;
			}
			if (count > maxSurplus) {
				maxSurplus = count;
				maxPt = pt;
			}
		}
		
		if (nearEmpty === null) {
			//done initially growing
			this.growing = false;
			return;
		}
		
		var size = nearEmpty.mineSize;
		if (numEmpty === 1) size *= 0.5;
		if (maxSurplus >= size) {
			this.MovePixels(maxPt,nearEmpty,maxSurplus);
		}
		return;
	}

	//determine cell radius from next empty mine
	var nMines = mapMines.length,
		i = 0, maxMineSz = 0;
	for (i = 0; i < nMines; i++) {
		var pt = mineOrder[i],
			count = pt.mineSize-pt.factionCount[facId];
		if (count > 0) break;
		maxMineSz = Math.max(maxMineSz, pt.mineSize);
	}
	if (i >= nMines) return;
	
	if (i !== this.lastOrderIndex) {
		this.extension = 0;
		this.lastOrderIndex = i;
	}
	
	//find surplus inside cell and enumerate wall points
	var pfLens = this.spawn.pointLengths,
		wallRad = pfLens[mineOrder[Math.max(0,i-1)].pointId]+this.extension,
		nPts = mapPoints.length,
		maxSurplus = 0, maxPt = null,
		allFull = true;
	for (i = 0; i < nPts; i++) {
		var pRad = pfLens[i]-wallRad,
			pt = mapPoints[i],
			measure = false;	
		if (pRad <= 0) {
			//inside cell, measure surplus and max mine size
			measure = true;
			
		} else {
			//outside cell, check if border wall point
			var cons = pt.connections,
				nCons = cons.length, k;
			for (k = 0; k < nCons; k++) {
				var ci = cons[k],
					cRad = pfLens[ci]-wallRad;
				if (cRad <= 0) break;
			}
			if (k < nCons) {
				abuf.push(pt);
				if (pt.factionCount[facId] < pt.mineSize) {
					allFull = false;
				}
			} else {
				measure = true;
			}
		}
		if (measure) {
			var count = pt.factionCount[facId]-pt.mineSize;
			if (count > maxSurplus) {
				maxSurplus = count;
				maxPt = pt;
			}
		}
	}
	
	var alen = abuf.length;
	if (allFull) {
		//check if cell wall full
		var cc = 0;
		for (i = 0; i < alen; i++) {
			cc += abuf[i].factionCount[facId];
		}
	
		//if mines full and cell wall full, extend wall
		if ((cc/alen) >= maxMineSz) {
			this.extension += 10;
		}
	}
	
	if (maxSurplus > 0 && alen > 0) {
		//pick move target from cell wall based on distance and fullness
		var total = 0, maxLen = maxPt.maxPointLength;
		pfLens = maxPt.pointLengths;
		for (i = 0; i < alen; i++) {
			var apt = abuf[i],
				prob = (maxLen-pfLens[apt.pointId])/(1+Math.max(0,apt.factionCount[facId]-apt.mineSize));
			bbuf.push(prob);
			total += prob;
		}
		
		if (total > 0) {
			var rind = this.RandFloat(0,total-1e-6),
				t = 0;
			for (i = 0; i < alen; i++) {
				var nxt = t+bbuf[i];
				if (rind < nxt) break;
				t = nxt;
			}
			this.MovePixels(maxPt,abuf[i], maxSurplus);
		}
		bbuf.length = 0;
	}
	abuf.length = 0;
}



/**@constructor*/
function ChronosMove() {
	this.src = null;
	this.dst = null;
	this.amount = 0;
}

/**@constructor
@extends {AI}*/
function Chronos(pId,seed) {
	AI.call(this,pId,seed);
	this.movePool = [];
}
Chronos.prototype = Object.create(AI.prototype);

Chronos.prototype.AddMove = function(src,dst,amount) {
	var pool = this.movePool,
		plen = pool.length,
		mo = null;
	if (plen) {
		mo = pool[--plen];
		pool.length = plen;
	} else {
		mo = new ChronosMove();
	}
	mo.src = src;
	mo.dst = dst;
	mo.amount = amount;
	abuf.push(mo);
}

Chronos.prototype.ClearMoves = function() {
	var alen = abuf.length, pool = this.movePool;
	for (var i = 0; i < alen; i++) pool.push(abuf[i]);
	abuf.length = 0;
}

Chronos.prototype.Decide = function() {
	//measure pixel production rate from all owned mines
	var nMines = mapMines.length,
		nPts = mapPoints.length,
		facId = this.faction,
		pixelProduction = 0, numOwned = 0;
	for (var i = 0; i < nMines; i++) {
		var mine = mapMines[i],
			sz = mine.mineSize;
		if (mine.factionCount[facId] >= sz) {
			pixelProduction += sz;
			numOwned++;
		}
	}
	if (pixelProduction) pixelProduction = 50/pixelProduction;//1/pixels produced per second
	else pixelProduction = 100;

	//if all mines owned, evenly distribute between points
	if (numOwned >= nMines) {
		var maxSurp = 0, maxCount = 0, maxPt = null;
		for (var i = 0; i < nPts; i++) {
			var p = mapPoints[i],
				c = p.factionCount[facId],
				s = p.mineSize;
			if (s > maxSurp) {
				maxSurp = s;
				maxCount = c;
				maxPt = p;
			}
		}
		if (maxSurp > 0) {
			var cons = maxPt.connections,
				nCons = cons.length,
				maxDiff = 0, maxCpt = null;
			for (var i = 0; i < nCons; i++) {
				var cpt = mapPoints[cons[i]],
					c = cpt.factionCount[facId]+cpt.travelCount[facId],
					df = maxCount-c;
				if (df > maxDiff) {
					maxDiff = df;
					maxCpt = cpt;
				}
			}
			if (maxDiff > 1) {
				this.MovePixels(maxPt,maxCpt, Math.min(maxSurp,Math.ceil(maxDiff*0.5)));
			}
		}
		return;
	}

	//search through all possible moves for best move
	var maxAdvantage = 0;
	for (var i = 0; i < nMines; i++) {
		var srcPt = mapMines[i],
			srcCount = srcPt.factionCount[facId],
			srcMine = srcPt.mineSize,
			srcSurplus = srcCount-srcMine;
		if (srcSurplus > 0) {
			//measure advantage given by moves
			var srcLens = srcPt.pointLengths;
			for (var m = 0; m < nMines; m++) {
				if (i === m) continue;
				
				var dstPt = mapMines[m],
					dstId = dstPt.pointId,
					dstMine = dstPt.mineSize,
					dstCount = dstMine-dstPt.factionCount[facId]-dstPt.travelCount[facId],
					dstEnemies = 0,
					sPt = srcPt;
				if (dstCount <= 0) continue;
					
				//count enemies on path to dst
				while (sPt !== dstPt) {
					var nxt = mapPoints[sPt.connections[sPt.findPath[dstId]]],
						fCount = nxt.factionCount,
						tCount = nxt.travelCount;
					for (var f = 0; f < MAX_FACTIONS; f++) {
						if (f === facId) continue;
						dstEnemies += fCount[f]+tCount[f];
					}
					sPt = nxt;
				}
				
				//calculate reward/cost
				var reward = dstMine, cost = dstEnemies,
					moveAmount = Math.min(srcSurplus, dstCount);
				cost = cost*pixelProduction+srcLens[dstId]/20;
				
				var advantage = reward/(1+cost);
				if (Math.abs(advantage-maxAdvantage) < 1e-8) {
					this.AddMove(srcPt,dstPt,moveAmount);
					
				} else if (advantage > maxAdvantage) {
					maxAdvantage = advantage;
					this.ClearMoves();
					this.AddMove(srcPt,dstPt,moveAmount);
				}
			}
		}
	}

	var movesLen = abuf.length;
	if (movesLen) {
		//if more than 1 best move, randomly select
		var mv = (movesLen === 1 ? abuf[0] : abuf[this.Rand(0,movesLen)]);
		this.MovePixels(mv.src,mv.dst, mv.amount);
		this.ClearMoves();
	}
}


