//Ethan Alexander Shulman 2025

var currentTime, deltaTime, tickTimer = 0, unpaused = true,
renderingEnabled = true, deathFxTick = false,
finaleTimer = 0, finaleFaded = false, finaleId = 0,
appStatus = -1;

const APPSTATUS_LOADING = -1,
APPSTATUS_MAINMENU = 0,
APPSTATUS_GAME = 1,
APPSTATUS_GAMEOVER = 2,
APPSTATUS_FINALE = 3,

TICK_RATE = 10, TICK_STEP = 1/TICK_RATE,

FINALE_PART_LENGTH = 10,
FINALE_TEXTS = ["Victory. Congratulations you have completed the campaign.",
"You have won the War of Pixels.</br>Continue playing in Quick Play and Custom Games.",
"Game created by Ethan Alexander Shulman.</br>XALOEZ.com"];



//DOM loaded, program entry point
document.addEventListener("DOMContentLoaded",function() {
	//initialize systems
	if (!InitializeGraphics()) return;
	InitializeRendering();
	InitializePixels();
	InitializeUI();
	InitializeMap();
	InitializeInput();
	InitializeAI();
	
	IndexedStorage_Open("WarOfPixels", function() {
		IndexedStorage_Load("SaveState", function(id, dat) {
			//load save state
			if (dat) {
				dat = JSON.parse(dat);
				playerFactionId = dat["f"];
				campaignTimes = dat["c"];
				musicVolume.value = dat["m"];
				deathEffectQuality.value = dat["d"];
				UpdateDeathFxQuality(parseInt(deathEffectQuality.value));
				quickPlayLevel = dat["q"];
				btnsMuted = dat["b"];
				UpdateMuteBtn(DGE("MuteBtnsBtn"));
				
			} else {
				//no save state, set default
				UpdateDeathFxQuality(2);
			}
			
			var nLvls = campaignTimes.length+1;
			for (var i = 0; i < nLvls; i++) UnlockCampaignLevel(i, true);
			
			UpdateVolume();
			UpdatePlayerFaction();
			
			FinishLoading();
		});
	});
});


//finish loading, close loading screen open main panel
function FinishLoading() {
	EV(loadingPanel,false);
	EV(mainMenuPanel,true);
	appStatus = APPSTATUS_MAINMENU;
	
	currentTime = Date.now();
	requestAnimationFrame(MainLoop);
}


//save game data state
function SaveGame() {
	if (DEV_VERSION) return;
	IndexedStorage_Store("SaveState", JSON.stringify(
	{"f":playerFactionId, "c":campaignTimes, "m":musicVolume.value,
	"d":deathFxQuality, "q":quickPlayLevel, "b":btnsMuted}));
}


//main update and rendering loop
function MainLoop() {
	var newTime = Date.now();
	deltaTime = (newTime-currentTime)/1000;
	currentTime = newTime;
	
	if (appStatus >= APPSTATUS_GAME && unpaused) {
		UpdateGameMusic();
		
		if (deathFxTick) {
			//delay death fx tick until frame after logic tick to split workload between frames
			if (deathFxQuality) UpdateDeathFx();
			deathFxTick = false;
		}
	
		//fixed tick rate for game logic/physics
		tickTimer = Math.min(1, tickTimer+deltaTime);
		while (tickTimer >= TICK_STEP) {
			if (!inTutorial) UpdateAI();
			UpdatePixels();
			UpdateMap();
			deathFxTick = true;
			
			tickTimer -= TICK_STEP;
		}
		gameTime += deltaTime;
	}
	if (renderingEnabled) {
		Render();
	}
	
	requestAnimationFrame(MainLoop);
}



//update finale
function UpdateFinale() {
	finaleTimer += deltaTime;
	if (!finaleFaded && finaleTimer >= FINALE_PART_LENGTH-1.5) {
		finaleText.style.animation = "1s FadeOut forwards";
		finaleFaded = true;
	}
	if (finaleTimer >= FINALE_PART_LENGTH) {
		finaleFaded = false;
		if (finaleId >= FINALE_TEXTS.length) {
			//done finale
			appStatus = APPSTATUS_GAMEOVER;
			EV(finalePanel,false);
			EV(gameOverPanel,true);
			return;
		}
	
		var ppx = playerFaction.pixels,
		targPx = ppx[Math.floor(Math.random()*ppx.length)];
		cameraX = targPx.pX;
		cameraY = targPx.pY;
		cameraZoom = 150+Math.round(Math.pow(finaleId,2)*100);
		finaleText.innerHTML = FINALE_TEXTS[finaleId++];
		finaleText.style.animation = "1s FadeIn forwards";
		
		finaleTimer -= FINALE_PART_LENGTH;
	}
	requestAnimationFrame(UpdateFinale);
}


