//Ethan Alexander Shulman 2025

var gameTime = 0, playingCampaignLevel = -1,
campaignTimes = [], quickPlayLevel = 0;

const INITIAL_PIXEL_COUNT = 30;


//load and launch game with settings, 'aiSettings' format [[aiIndex, aiDelay, aiSeed]]
function LaunchGame(mapSize, mapSeed, aiSettings, aiColors) {
	SetFactions(aiSettings.length, aiColors);
	GenerateMap(mapSize,mapSeed);
	GeneratePathFinding();
	LoadAI(aiSettings);
	
	for (var i = 0; i < mapFactionSpawns.length; i++) {
		//spawn initial pixels for each faction
		var fac = activeFactions[i],
			pt = mapFactionSpawns[i];
		for (var k = 0; k < INITIAL_PIXEL_COUNT; k++) fac.Spawn(pt);
	}
	
	var pspawn = mapFactionSpawns[0];
	cameraX = pspawn.x;
	cameraY = pspawn.y;
	cameraZoom = 256;
	
	DelayNextSong();
}

function FinishLaunchingGame() {
	gameTime = 0;
	currentTime = Date.now();
	tickTimer = 0;
	appStatus = APPSTATUS_GAME;
	EV((inTutorial ? tutorialPanel : gamePanel), true);
}


//game finished
function GameOver(playerWon, tie) {
	//for game over screen send all remaining pixels to center point and focus camera on it
	var nPts = mapPoints.length,
		center = mapPoints[nPts<=16 ? 6 : (nPts<=64 ? 24 : 68)];
	cameraX = center.x;
	cameraY = center.y;
	cameraZoom = 512;
	for (var f = 0; f < activeFactions.length; f++) {
		PixelsTravelTarget(activeFactions[f].pixels, center);
	}

	//display game over screen
	DGE("GameResult").textContent = (tie ? "Tie" : (playerWon ? "Win" : "Loss"));
	
	//update quick play level
	if (!tie) {
		if (playerWon) quickPlayLevel = Math.max(quickPlayLevel, currentMaxAiIndex+1);
		else quickPlayLevel = Math.min(quickPlayLevel, Math.max(0, currentMaxAiIndex-1));
	}
	
	var timeSec = Math.round(gameTime),
		newFastest = false, newLevel = false;
	if (playingCampaignLevel > -1) {
		if (campaignTimes.length > playingCampaignLevel) {
			var oldTime = campaignTimes[playingCampaignLevel];
			if (gameTime < oldTime) {
				campaignTimes[playingCampaignLevel] = gameTime;
				newFastest = true;
			}
			
		} else {
			campaignTimes.push(gameTime);
			newFastest = true;
			if (playingCampaignLevel < CAMPAIGN_LEVELS.length-1) {
				newLevel = true;
				UnlockCampaignLevel(playingCampaignLevel+1, true);
			}
		}
		
		playingCampaignLevel = -1;
	}
	DGE("GameTime").innerHTML = TimeString(timeSec) + (newFastest ? ", new fastest time!" : "")
								+ (newLevel ? "</br>New level unlocked." : "");


	SaveGame();
	EV(gamePanel,false);
	EV(pausePanel,false);
	if (playingCampaignLevel === 99) {
		finaleId = 0;
		finaleTimer = FINALE_PART_LENGTH;
		
		appStatus = APPSTATUS_FINALE;
		EV(finalePanel,true);
		UpdateFinale();
	
	} else {
		appStatus = APPSTATUS_GAMEOVER;
		EV(gameOverPanel,true);
	}
}


//quit game back to main menu
function QuitGame() {
	ClearMap();
	ClearPixels();
	
	cameraX = 1024;
	cameraY = 1024;
	cameraZoom = 256;
	unpaused = true;
	appStatus = APPSTATUS_MAINMENU;

	EV(gamePanel,false);
	EV(pausePanel,false);
	EV(mainMenuPanel,true);
}


//launch quick play game
function QuickPlay() {
	var mapSize = Math.floor(Math.random()*3),
		aiCount = 1+Math.floor(Math.random()*(mapSize === 2 ? 5 : 3)),
		ais = new Array(aiCount);
	for (var i = 0; i < aiCount; i++) {
		var ind = Math.floor(Math.min(AI_INFO.length, Math.random()*(quickPlayLevel+2)));
		ais[i] = [ind, 0.1+Math.pow(Math.random(),2)*9.9, Math.floor(Math.random()*0xFFFFFFFF)];
	}
	LaunchGame(mapSize, Math.floor(Math.random()*0xFFFFFFFF), ais, Math.floor(Math.random()*5));
}


const CAMPAIGN_LEVELS = [
//format [mapSize, mapSeed, aiColorShift, [[aiIndex, aiDelay, aiSeed]]]

//#1 - small, doofus 60s
[0, 3083470281, 0, [[0, 60, 380164950]]],
//#2 - small, doofus 1s
[0, 1739154609, 0, [[0, 1, 1601284859]]],
//#3 - small, doofus 8s
[0, 2452807339, 0, [[0, 8, 2452807339]]],
//#4 - small, doofus 1s, doofus 3s
[0, 387353626, 1, [[0, 1, 1379662218], [0, 3, 3682410399]]],
//#5 - small, doofus 1s, plushkin 1.5s
[0, 1030761191, 1, [[0, 1, 4010175939], [1, 1.5, 485232231]]],
//#6 - small, plushkin 1s, plushkin 4.2s
[0, 4195110482, 1, [[1, 1, 2688133161], [1, 4.2, 1984874534]]],
//#7 - medium, doofus 6.5s, mus 1.5s, unimodus 5s
[1, 1911934147, 0, [[0, 6.5, 2635119771], [2, 1.5, 3556072636], [3, 5, 1894044780]]],
//#8 - small, mus 1s
[0, 3593261793, 1, [[2, 1, 1866941765]]],
//#9 - small, voidling 1s
[0, 1356328326, 2, [[4, 1, 1827309136]]],
//#10 - medium, mus 0.5s, unimodus 3.5s, voidling 1s
[1, 3622434084, 1, [[2, 0.5, 3139944499], [3, 3.5, 1748981798], [4, 1, 2047297224]]],

//#11 - medium, unimodus 0.5s, unimodus 8s
[1, 2344043256, 2, [[3, 0.5, 2364441160], [3, 8, 2364441160]]],
//#12 - medium, doofus 5s, voidling 0.8s, plushkin 3s, mus 1s, unimodus 1.5s
[1, 246469632, 0, [[0, 5, 3451173645], [4, 0.8, 540028157], [1, 3, 1106654210], [2, 1, 668445274], [3, 1.5, 3440734088]]],
//#13 - small, voidling 0.5s, voidling 4s
[0, 3666142333, 2, [[4, 0.5, 2785723447], [4, 4, 1518008710]]],
//#14 - medium, voidguard 1.2s
[1, 894814026, 4, [[5, 1.2, 576612916]]],
//#15 - small, voidling 0.5s, voidling 4s, voidling 15s
[0, 3084808287, 0, [[4, 0.5, 1426408271], [4, 4, 1343424416], [4, 15, 2302218742]]],
//#16 - medium, voidling 1.5s, unimodus 0.5s, unimodus 15s, voidguard 0.5s
[1, 2599846498, 1, [[4, 1.5, 2128698502], [3, 0.5, 3727644477], [3, 15, 1385784903], [5, 0.5, 3389037480]]],
//#17 - medium, voidguard 1s, voidguard 5s, voidguard 12s
[1, 3492188185, 1, [[5, 1, 3459297027], [5, 5, 825261909], [5, 12, 1458669230]]],
//#18 - medium, unimodus 2s
[1, 1973448989, 0, [[3, 2, 3332972174]]],
//#19 - small, voidguard 3s
[0, 339127466, 3, [[5, 3, 2879996479]]],
//#20 - large, striker 6.2s, unimodus 1.5s, mus 5s
[2, 685679639, 1, [[6, 6.2, 3812633119], [3, 1.5, 772225259], [2, 5, 992993774]]],

//#21 - small, striker 0.6s
[0, 356304632, 4, [[6, 0.6, 2899404516]]],
//#22 - medium, voidguard 0.4s, voidling 4.5s, voidguard 6s, voidling 7s, striker 0.5s
[1, 969294080, 0, [[5, 0.4, 2694921834], [4, 4.5, 4082697939], [5, 6, 1865812089], [4, 7, 2077298127], [6, 0.5, 1944587469]]],
//#23 - medium, striker 0.4s, striker 7s, striker 0.5s
[1, 1382279964, 1, [[6, 0.4, 2523523421], [6, 7, 2569952441], [6, 0.5, 1252526724]]],
//#24 - medium, striker 1.5s
[1, 3543616048, 2, [[6, 1.5, 312267440]]],
//#25 - large, striker 0.7s, voidguard 2s, globuulus 4s, unimodus 5s
[2, 1544064826, 0, [[6, 0.7, 215304324], [5, 2, 3982063482], [7, 4, 896469929], [3, 5, 2457787539]]],
//#26 - small, globuulus 1.5s
[0, 2048455609, 1, [[7, 1.5, 2327887556]]],
//#27 - medium, doofus 10s, globuulus 1.2s, plushkin 2s
[1, 3430612088, 0, [[0, 10, 1805018713], [7, 1.2, 1805018713], [1, 2, 1667103317]]],
//#28 - medium, mus 2s, doofus 4.4s, striker 1.3s, doofus 0.5s, mus 8s
[1, 3300494461, 0, [[2, 2, 169837387], [0, 4.4, 215385492], [6, 1.3, 1718063738], [0, 0.5, 410830354], [2, 8, 3817271792]]],
//#29 - medium, globuulus 2s, globuulus 0.5s, globuulus 2.3s
[1, 1621106092, 3, [[7, 2, 2814416260], [7, 0.5, 2654421772], [7, 2.3, 952749805]]],
//#30 - large, adad 3s
[2, 1996761831, 2, [[8, 3, 2694310665]]],

//#31 - medium, voidguard 1.4s, adad 4.5s, striker 0.5s
[1, 1572667069, 1, [[5, 1.4, 3028010164], [8, 4.5, 2102046504], [6, 0.5, 2483748025]]],
//#32 - medium, voidling 0.6s, globuulus 10s
[1, 2568136833, 2, [[4, 0.6, 4237552241], [7, 10, 1578388086]]],
//#33 - large, mus 8s, voidguard 3.2s, adad 1s, striker 1.2s, mus 0.5s
[2, 125615636, 0, [[2, 8, 3914747106], [5, 3.2, 1348597116], [8, 1, 1390096855], [6, 1.2, 2354839325], [2, 0.5, 3428129316]]],
//#34 - medium, voidling 4.5s, voidguard 2s, unimodus 10s
[1, 811795389, 2, [[4, 4.5, 3380220541], [5, 2, 833296273], [3, 10, 2784478393]]],
//#35 - small, adad 1.2s
[0, 776510710, 2, [[8, 1.2, 3633939271]]],
//#36 - medium, adad 3s
[1, 1546503130, 2, [[8, 3, 2112429249]]],
//#37 - large, globuulus 0.4s, voidguard 0.5s, doofus 8s
[2, 1839425697, 1, [[7, 0.4, 3593338251], [5, 0.5, 3955135647], [0, 8, 2957807387]]],
//#38 - medium, unimodus 8s, adad 0.4s
[1, 449051727, 0, [[3, 8, 3922911463], [8, 0.4, 3475705463]]],
//#39 - small, voidguard 1s, voidguard 1.5s, adad 5s
[0, 4158862241, 0, [[5, 1, 1040450061], [5, 1.5, 3385322495], [8, 5, 2051638554]]],
//#40 - medium, voidlord 1s
[1, 711797055, 3, [[9, 1, 3843479461]]],

//#41 - large, voidlord 3s, plushkin 15s, doofus 10s, voidguard 8s
[2, 2843146849, 0, [[9, 3, 1901599990], [1, 15, 2279959377], [0, 10, 722049123], [5, 8, 4235033553]]],
//#42 - medium, voidling 0.2s, voidlord 2s
[1, 3733182711, 2, [[4, 0.2, 2854765572], [9, 2, 2982500234]]],
//#43 - large, adad 0.5s, striker 4s
[2, 1615597379, 1, [[8, 0.5, 241091222], [6, 4, 1701471766]]],
//#44 - medium, adad 2s, voidlord 4s
[1, 2637059893, 3, [[8, 2, 2518842578], [9, 4, 694310468]]],
//#45 - small, voidlord 0.5s
[0, 4079882391, 2, [[9, 0.5, 957728198]]],
//#46 - medium, adad 1s, adad 3s, adad 10s
[1, 3696103468, 1, [[8, 1, 2088187697], [8, 3, 4009682588], [8, 10, 1721020903]]],
//#47 - medium, unimodus 10s, voidlord 2s
[1, 2643559677, 3, [[3, 10, 3839254853], [9, 2, 3111244671]]],
//#48 - large, voidling 2.5s, voidlord 3s, voidling 5s, voidling 8s, voidling 1.5s
[2, 1650632319, 4, [[4, 2.5, 646670331], [9, 3, 331281582], [4, 5, 2381729136], [4, 8, 3591821967], [4, 1.5, 94683324]]],
//#49 - small, voidguard 2s, voidlord 1s, voidguard 0.4s
[0, 3333963817, 0, [[5, 2, 3730025017], [9, 1, 2839854946], [5, 0.4, 3973616950]]],
//#50 - large, voidlord 5s
[2, 125784197, 3, [[9, 5, 642932863]]],

//#51 - large, globuulus 8s, voidlord 3s, adad 0.4s
[2, 2349394766, 1, [[7, 8, 2905849358], [9, 3, 3693074425], [8, 0.4, 205655290]]],
//#52 - medium, striker 1.1s, mus 8s, voidguard 3s, doofus 5s, voidling 4.6s
[1, 942950086, 0, [[6, 1.1, 173139435], [2, 8, 2404279259], [5, 3, 960776826], [0, 5, 4075264648], [4, 4.6, 2185436819]]],
//#53 - small, adad 2s, voidlord 2.5s
[0, 1616332321, 2, [[8, 2, 3057678503], [9, 2.5, 3310000677]]],
//#54 - medium, voidlord 0.4s, voidlord 6s, voidlord 2.5s
[1, 2877584474, 0, [[9, 0.4, 2989964105], [9, 6, 2587196678], [9, 2.5, 3061223228]]],
//#55 - large, leviathan 2s
[2, 4018183046, 4, [[10, 2, 3711964936]]],
//#56 - medium, unimodus 3s, voidling 1.4s, leviathan 1.5s, voidguard 4s
[1, 4057903493, 3, [[3, 3, 3658995852], [4, 1.4, 1630894387], [10, 1.5, 1864539559], [5, 4, 723148873]]],
//#57 - medium, voidlord 8s, adad 3.4s, adad 3.5s, globuulus 2s
[1, 37265294, 1, [[9, 8, 1952591543], [8, 3.4, 450884315], [8, 3.5, 3598276165], [7, 2, 3961357919]]],
//#58 - large, striker 3.4s, leviathan 2.3s, voidling 4s, unimodus 5s, voidguard 2.6s
[2, 1153071357, 4, [[6, 3.4, 449543012], [10, 2.3, 1729269362], [4, 4, 3151614347], [3, 5, 4139277907], [5, 2.6, 1985167989]]],
//#59 - large, voidlord 7s, adad 0.9s, adad 5s, voidlord 3s
[2, 3031769704, 0, [[9, 7, 3664879664], [8, 0.9, 3896089869], [8, 5, 308663872], [9, 3, 2024328965]]],
//#60 - small, leviathan 0.6s
[0, 2784997789, 3, [[10, 0.6, 4133002373]]],

//#61 - medium, adad 2.5s, voidlord 3s, leviathan 1.4s
[1, 3229812069, 2, [[8, 2.5, 4247683478], [9, 3, 397772372], [10, 1.4, 147701700]]],
//#62 - medium, unimodus 8s, voidlord 1.2s, unimodus 5s, voidling 2s, unimodus 3s
[1, 2738723703, 0, [[3, 8, 1019017518], [9, 1.2, 2129736130], [3, 5, 508243639], [4, 2, 2368264666], [3, 3, 2324850739]]],
//#63 - medium, leviathan 5s
[1, 1961532134, 4, [[10, 5, 874241428]]],
//#64 - small, cell 5s, leviathan 2s, voidlord 3s
[0, 2828346832, 2, [[11, 5, 2697202295], [10, 2, 671006303], [9, 3, 26027277]]],
//#65 - large, voidlord 1.2s, leviathan 1.5s, voidlord 2.5s, adad 3s
[2, 2505257317, 1, [[9, 1.2, 1632246762], [10, 1.5, 1473349959], [9, 2.5, 3282950633], [8, 3, 3628204183]]],
//#66 - medium, cell 3s
[1, 3408883437, 0, [[11, 3, 1876916006]]],
//#67 - large, leviathan 2s, cell 5s, adad 3.4s, cell 4s, leviathan 5.4s
[2, 838761204, 3, [[10, 2, 3158540228], [11, 5, 507894567], [8, 3.4, 4001952408], [11, 4, 339291295], [10, 5.4, 3526591756]]],
//#68 - small, cell 0.4s
[0, 3535417971, 1, [[11, 0.4, 1139187542]]],
//#69 - large, voidlord 10s, voidlord 4s, voidlord 0.6s
[2, 987713670, 2, [[9, 10, 501830857], [9, 4, 1446396641], [9, 0.6, 3197288744]]],
//#70 - large, cell 1s
[2, 481100907, 3, [[11, 1, 2712120838]]],

//#71 - medium, leviathan 3s, leviathan 0.6s, cell 7s, leviathan 2.6s
[1, 840915741, 0, [[10, 3, 3216355415], [10, 0.6, 1675319260], [11, 7, 1705000819], [10, 2.6, 545395885]]],
//#72 - medium, adad 2.4s, cell 0.5s
[1, 2932720570, 2, [[8, 2.4, 13110676], [11, 0.5, 3525725296]]],
//#73 - small, ra 1.6s
[0, 4257972317, 3, [[12, 1.6, 4178428844]]],
//#74 - small, cell 1s, ra 7s
[0, 1840603858, 1, [[11, 1, 3117707866], [12, 7, 1797761918]]],
//#75 - large, cell 0.6s, leviathan 8s
[2, 2009859054, 0, [[11, 0.6, 389839639], [10, 8, 895924394]]],
//#76 - large, adad 3s, ra 4.5s
[2, 373373032, 4, [[8, 3, 3390978513], [12, 4.5, 1173930803]]],
//#77 - medium, ra 2s
[1, 1808084210, 4, [[12, 2, 570763836]]],
//#78 - medium, voidlord 0.6s, ra 4s
[1, 454455461, 1, [[9, 0.6, 333456981], [12, 4, 1602813237]]],
//#79 - medium, adad 0.5s, leviathan 0.8s, voidlord 1.1s, cell 7s
[1, 1240437262, 0, [[8, 0.5, 121196066], [10, 0.8, 4146747763], [9, 1.1, 1948153095], [11, 7, 2607555632]]],
//#80 - large, ra 0.8s
[2, 883275259, 3, [[12, 0.8, 976521823]]],

//#81 - large, striker 1.2s, cell 4s, achilles 3.4s, cell 0.6s, leviathan 8s
[2, 2106119887, 1, [[6, 1.2, 2248051966], [11, 4, 3305269442], [13, 3.4, 4112260762], [11, 0.6, 2127523485], [10, 8,  2223003678]]],
//#82 - medium, achilles 5s
[1, 2296998650, 4, [[13, 5, 1459021621]]],
//#83 - small, achilles 12s, adad 3s, unimodus 0.8s
[0, 1628737270, 2, [[13, 12, 1166539627], [8, 3, 2565033336], [3, 0.8, 3139307212]]],
//#84 - small, achilles 2s
[0, 2310223700, 3, [[13, 2, 634206070]]],
//#85 - medium, cell 3s, achilles 5s, leviathan 0.8s, cell 2s
[1, 572064836, 4, [[11, 3, 3007154537], [13, 5, 2816897644], [10, 0.8, 3138052446], [11, 2, 428768266]]],
//#86 - large, cell 1.4s, cell 7s, cell 2.1s, ra 5.4s, adad 0.6s
[2, 4179293878, 0, [[11, 1.4, 199580256], [11, 7, 2338579500], [11, 2.1, 4155742957], [12, 5.4, 1598825030], [8, 0.6, 3408355014]]],
//#87 - large, achilles 0.6s
[2, 2872536190, 2, [[13, 0.6, 3982696729]]],
//#88 - medium, ra 3.4s, kali 11s
[1, 2002972257, 1, [[12, 3.4, 931685722], [14, 11, 410996913]]],
//#89 - medium, ra 8s, ra, 1.6s, ra 3.5s, ra 4.3s
[1, 3141057868, 0, [[12, 8, 4104870563], [12, 1.6, 4184760165], [12, 3.5, 2081814243], [12, 4.3, 388183541]]],
//#90 - small, kali 4s
[0, 1700058376, 3, [[14, 4, 1524764050]]],

//#91 - medium, kali 0.8s  973306110
[1, 2962938291, 4, [[14, 0.8, 861068966]]],
//#92 - large, kali, 1s
[2, 986616330, 1, [[14, 1, 3907326076]]],
//#93 - medium, achilles 1.6s, kali 2s, ra 1.1s
[1, 2880867678, 0, [[13, 1.6, 2759956942], [14, 2, 2942284096], [12, 1.1, 1257754242]]],
//#94 - large, cell 0.7s, cell 8s, chronos 3s, cell 10s, cell 1.1s
[2, 1721671792, 1, [[11, 0.7, 2759751782], [11, 8, 2761319824], [15, 3, 79541925], [11, 10, 3885663374], [11, 1.1, 3125089458]]],
//#95 - small, chronos 1.2s
[0, 3153966027, 2, [[15, 1.2, 2143676668]]],
//#96 - small, kali 5s, achilles 1.6s, chronos 3s
[0, 1551171767, 3, [[14, 5, 95401412], [13, 1.6, 1237400488], [15, 3, 3572691760]]],
//#97 - medium, chronos 1.8s
[1, 3860647534, 4, [[15, 1.8, 470958788]]],
//#98 - medium, ra 1.4s, kali 3.5s, chronos 2.5s, achilles 2.2s
[1, 484741188, 0, [[12, 1.4, 270335280], [14, 3.5, 3456365302], [15, 2.5, 2639257615], [13, 2.2, 2003842632]]],
//#99 - large, chronos 1s
[2, 1114199563, 3, [[15, 1, 3248580533]]],
//#100 - large, chronos 1.6s, kali 1.1s, chronos 0.5s, kali 3s, chronos 2.4s
[2, 153164463, 1, [[15, 1.6, 1231140836], [14, 1.1, 704840495], [15, 0.5, 2144401266], [14, 3, 1726315622], [15, 2.4, 3873319422]]]
];
