//Ethan Alexander Shulman 2024

var draggingCamera = false, touchPanLastX = 0, touchPanLastY = 0, touchZoomLast = 0,
selectingPixels = false, selectBoxLeft = 0, selectBoxBottom = 0, selectCenterX = 0, selectCenterY = 0, selectExtentX = 0, selectExtentY = 0,
targetPoint = null,
usingTouchInput = false;

const TARGET_POINT_RANGE = 40;


function InitializeInput() {
	//mouse input
	window.addEventListener("mousedown", function(e) {
		if (appStatus === APPSTATUS_GAME) {
			if (e.button === 0) {
				draggingCamera = true;
			
			} else if (e.button === 2) {
				selectingPixels = true;
				selectBoxLeft = ScreenToWorld(e.clientX*pixelScaling, e.clientY*pixelScaling);
				selectBoxBottom = returnY;
				selectExtentX = 0;
				selectExtentY = 0;
			
				e.preventDefault();
			}
		}
	});
	window.addEventListener("mouseup", function(e) {
		if (appStatus === APPSTATUS_GAME) {
			if (e.button === 0) {
				draggingCamera = false;
			
			} else if (e.button === 2) {
				selectingPixels = false;
				UpdateSelectBox(e.clientX*pixelScaling, e.clientY*pixelScaling);
				FinishSelectBox();
			
				e.preventDefault();
			}
		}
	});
	window.addEventListener("contextmenu", function(e) {
		e.preventDefault();
	});
	window.addEventListener("mousemove", function(e) {
		if (appStatus === APPSTATUS_GAME) {
			if (draggingCamera) {
				//camera drag pan
				var mx = e.movementX*pixelScaling,
					my = e.movementY*pixelScaling;
				cameraX = Math.max(0,Math.min(2048, cameraX-ScaleScreenToWorld(mx,my)));
				cameraY = Math.max(0,Math.min(2048, cameraY-returnY));
			}
			
			if (selectingPixels) {
				UpdateSelectBox(e.clientX*pixelScaling, e.clientY*pixelScaling);
			}
			
			if (selectedPixels.length) {
				//update nearest point target for moving selected pixels
				var wX = ScreenToWorld(e.clientX*pixelScaling, e.clientY*pixelScaling),
					wY = returnY,
					npt = NearestMapPoint(wX,wY);
				targetPoint = (returnY < TARGET_POINT_RANGE ? npt : null);
			}
		}
	});
	window.addEventListener("wheel", function(e) {
		//camera zoom
		if (appStatus === APPSTATUS_GAME) {
			var zscale;
			if (e.ctrlKey || !Number.isInteger(e.deltaY)) {
				zscale = e.deltaY;
				if (e.deltaMode === 1) zscale *= 40;//convert to approx pixels
				else if (e.deltaMode === 2) zscale *= 800;
				zscale = Math.pow(2, zscale*0.01);//pixels to zoom factor
			
			} else {
				zscale = (e.deltaY>0 ? 1.41421356237 : 0.70710678118);
			}
			cameraZoom = Math.max(16, Math.min(2048, Math.round(cameraZoom*zscale)));
		}
		if (e.ctrlKey) e.preventDefault();
	}, {"passive":false});
	
	
	//keyboard input
	window.addEventListener("keydown", function(e) {
		if (appStatus == APPSTATUS_GAME) {
			if (e.code === "Escape" || e.code === "Space") {
				Pause();
			
			} else if (e.code === "Digit1") {
				cameraZoom = Math.max(16, Math.round(cameraZoom/2));
			} else if (e.code === "Digit2") {
				cameraZoom = Math.min(2048, cameraZoom*2);
			}
		}
	});
	
	
	//touch input
	window.addEventListener("touchstart",function(e) {
		usingTouchInput = true;

		if (!(document.fullscreenElement || document.webkitFullscreenElement)) {//set mobile to fullscreen
			if (document.body.requestFullscreen) document.body.requestFullscreen();
			else if (document.body.webkitRequestFullscreen) document.body.webkitRequestFullscreen();
		}
		
		if (!(e.target === gamePanel || e.target === pausePanel || e.target === tutorialPanel)) return;
		
		if (appStatus === APPSTATUS_GAME) {
			var tch = e.touches,
				nTouches = tch.length;
			if (nTouches > 1) {
				//zoom/pan
				var x0 = tch[0].clientX, y0 = tch[0].clientY,
					x1 = tch[1].clientX, y1 = tch[1].clientY,
					dx = x1-x0, dy = y1-y0;
				touchPanLastX = x0+dx*0.5;
				touchPanLastY = y0+dy*0.5;
				touchZoomLast = Math.sqrt(dx*dx+dy*dy);
				selectingPixels = false;
			
			} else {
				//select box
				selectingPixels = true;
				selectBoxLeft = ScreenToWorld(tch[0].clientX*pixelScaling, tch[0].clientY*pixelScaling);
				selectBoxBottom = returnY;
				selectExtentX = 0;
				selectExtentY = 0;
			}
		}
		
		e.preventDefault();
	});
	window.addEventListener("touchmove",function(e) {
		if (!(e.target === gamePanel || e.target === pausePanel || e.target === tutorialPanel)) return;
	
		if (appStatus === APPSTATUS_GAME) {
			var tch = e.touches,
				nTouches = tch.length;
			if (nTouches > 1) {
				//zoom/pan
				var x0 = tch[0].clientX, y0 = tch[0].clientY,
					x1 = tch[1].clientX, y1 = tch[1].clientY,
					dx = x1-x0, dy = y1-y0,
					currPanX = x0+dx*0.5, currPanY = y0+dy*0.5,
					currZoom = Math.sqrt(dx*dx+dy*dy),
					screenPanX = currPanX-touchPanLastX, screenPanY = currPanY-touchPanLastY;
				if (Math.sqrt(screenPanX*screenPanX+screenPanY*screenPanY) > Math.abs(currZoom-touchZoomLast)) {
					var panX = ScaleScreenToWorld(screenPanX*pixelScaling, screenPanY*pixelScaling);
					cameraX = Math.max(0,Math.min(2048, cameraX-panX));
					cameraY = Math.max(0,Math.min(2048, cameraY-returnY));
					
				} else if (currZoom && touchZoomLast) {
					cameraZoom = Math.max(16, Math.min(2048, Math.round(cameraZoom*touchZoomLast/Math.max(1e-100,currZoom))));
				}
				
				touchPanLastX = currPanX;
				touchPanLastY = currPanY;
				touchZoomLast = currZoom;
			
			} else {
				//select box
				UpdateSelectBox(tch[0].clientX*pixelScaling, tch[0].clientY*pixelScaling);
			}
		}

		e.preventDefault();
	});
	var touchEnd = function(e) {
		if (!(e.target === gamePanel || e.target === pausePanel || e.target === tutorialPanel)) return;
	
		if (appStatus === APPSTATUS_GAME) {
			if (selectingPixels) {
				var tch = e.changedTouches;
				selectingPixels = false;
				UpdateSelectBox(tch[0].clientX*pixelScaling, tch[0].clientY*pixelScaling);
				FinishSelectBox();
			}
		}
		
		e.preventDefault();
	};
	window.addEventListener("touchend",touchEnd);
	window.addEventListener("touchcancel",touchEnd);
}


//update select box drag
function UpdateSelectBox(right,top) {
	var bright = ScreenToWorld(right,top),
		btop = returnY;
	selectCenterX = (selectBoxLeft+bright)*0.5;
	selectCenterY = (selectBoxBottom+btop)*0.5;
	selectExtentX = Math.abs(bright-selectBoxLeft)*0.5;
	selectExtentY = Math.abs(btop-selectBoxBottom)*0.5;
}


//select pixels of faction within box and store in dstArray
function SelectPixelsWithinBox(faction, dstArray, centerX,centerY,extX,extY) {
	var ppx = faction.pixels, facId = faction.id,
		pxLen = ppx.length,
		sExtX = extX+PIXEL_RADIUS,
		sExtY = extY+PIXEL_RADIUS;
	for (var i = 0; i < pxLen; i++) {
		var px = ppx[i],
			dX = Math.abs(px.x-centerX),
			dY = Math.abs(px.y-centerY);
		if (dX < extX && dY < extY) {
			if (!px.travelPoint) {
				var opt = px.orbitPoint;
				if (opt.selectedCount >= opt.factionCount[facId]-opt.mineSize) continue;
				opt.selectedCount++;
			}
			px.selected = true;
			dstArray.push(px);
		}
	}
}

//deselect pixels
function DeselectPixels() {
	for (var i = 0, l = selectedPixels.length; i < l; i++) {
		selectedPixels[i].selected = false;
	}
	for (var i = 0, l = mapPoints.length; i < l; i++) mapPoints[i].selectedCount = 0;
	selectedPixels.length = 0;
}


//finish select box
function FinishSelectBox() {
	if (selectedPixels.length) {
		//attempt to click target point
		var npt = NearestMapPoint(selectBoxLeft,selectBoxBottom);
		targetPoint = (returnY < TARGET_POINT_RANGE ? npt : null);
		if (targetPoint !== null) {
			//move selected pixels to target point
			PixelsTravelTarget(selectedPixels, targetPoint);
			DeselectPixels();
			UpdateSelectCount();
			targetPoint = null;
			return;
		}
	}

	//select pixels within box
	DeselectPixels();
	targetPoint = null;
	SelectPixelsWithinBox(playerFaction,selectedPixels, selectCenterX,selectCenterY,selectExtentX,selectExtentY);
	UpdateSelectCount();
}
