</script>

<script type="text/plain" id="BgVert.glsl">
//background vertex shader
attribute vec2 vPos;
varying vec4 uv;

uniform vec4 camera;//xy = camera position, zw = 1/(aspect*zoom)

void main() {
	vec2 wpos = vPos/camera.zw+camera.xy;
	#define GRID 50.
	uv = vec4(wpos/GRID, (wpos-1024.)*.0003);
	gl_Position = vec4(vPos.xy,0,1);
}
</script>

<script type="text/plain" id="Bg.glsl">
//background frag shader
varying vec4 uv;

void main() {
	vec2 lp = abs(uv.xy-floor(uv.xy)-.5);
	float gv = max(max(lp.x,lp.y)-length(uv.zw), 0.);
	gl_FragColor = vec4(vec3(.3*gv),1);
}
</script>


<script type="text/plain" id="SelectVert.glsl">
//select box vertex shader
attribute vec2 vPos;
varying vec4 uv;

uniform vec4 data[2];
#define CAM_POS (data[0].xy)
#define CAM_SCALE (data[0].zw)
#define BOX_POS (data[1].xy)
#define BOX_EXT (data[1].zw)

void main() {
	vec2 wpos = vPos*BOX_EXT;
	uv = vec4(wpos,BOX_EXT);
	gl_Position = vec4((wpos+BOX_POS-CAM_POS)*CAM_SCALE,0,1);
}
</script>

<script type="text/plain" id="Select.glsl">
//select box frag shader
varying vec4 uv;

void main() {
	vec2 bp = uv.zw-abs(uv.xy);
	float mx = min(bp.x,bp.y);
	gl_FragColor = vec4(.5/(1.+mx*.5));
}
</script>


<script type="text/plain" id="TargetVert.glsl">
//point target vertex shader
attribute vec2 vPos;
varying vec4 uv;

uniform vec4 data[2];
#define CAM_POS (data[0].xy)
#define CAM_SCALE (data[0].zw)
#define TARGET_POS (data[1].xy)
#define TARGET_EXT (data[1].zw)

void main() {
	vec2 wpos = vPos*TARGET_EXT;
	uv = vec4(wpos,TARGET_EXT);
	gl_Position = vec4((wpos+TARGET_POS-CAM_POS)*CAM_SCALE,0,1);
}
</script>

<script type="text/plain" id="Target.glsl">
//point target frag shader
varying vec4 uv;

#define PI 3.141592653

void main() {
	#define REPEAT (PI*2./8.)
	float rang = mod(atan(uv.x,uv.y)+PI, REPEAT)-REPEAT*.5;
	vec2 ap = vec2(sin(rang),cos(rang))*length(uv.xy);
	
	ap.x = abs(ap.x);
	ap.y = 6.-ap.y;
	float tri = max(dot(ap,normalize(vec2(1.5,1)))-1., .5-ap.y);
	gl_FragColor = vec4(.6*clamp(1.-tri*.25,0.,1.)/(1.+max(0.,tri)*4.));
}
</script>


<script type="text/plain" id="SpriteVert.glsl">
//sprite vertex shader
attribute vec4 vPos;
varying vec2 uv;

uniform vec4 camera;//xy = camera position, zw = 1/(aspect*zoom)

void main() {
	uv = vPos.zw;
	gl_Position = vec4((vPos.xy-camera.xy)*camera.zw,0,1);
}
</script>


<script type="text/plain" id="Path.glsl">
//map path line shader
varying vec2 uv;

void main() {
	vec2 p = abs(uv);
	p.x = max(0.,p.x-.1);
	gl_FragColor = vec4(.35*max(0.,1.-p.x*(1./.9)-clamp((p.y-.9)*10.,0.,1.))/(1.+p.x*2.5));
}
</script>


<script type="text/plain" id="Point.glsl">
//map point shader
varying vec2 uv;

void main() {
	float l = max(0.,length(uv)-.1);
	gl_FragColor = vec4(.35/(1.+l*4.)*max(0.,1.-l*(1./.9)));
}
</script>


<script type="text/plain" id="Mine.glsl">
//map mine shader
varying vec2 uv;

void main() {
	float l = length(uv),
		a = atan(uv.x,uv.y);
	gl_FragColor = vec4((sin(a*7.+l*6.28*4.)*.5+.5)*
					(l < .8 ? l*(1./.8)/(1.+(.8-l)*20.) : (1.-l)*(1./(1.-.8))));
}
</script>


<script type="text/plain" id="Pixel.glsl">
//pixel shader
varying vec2 uv;
uniform vec4 color;

void main() {
	vec2 p = abs(uv);
	float av = max(p.x,p.y)*1.5;
	gl_FragColor = color*(av < .5 ? 1. : (1.5-av)/(1.+(av-.5)*3.));
}
</script>


<script type="text/plain" id="WaveSimVert.glsl">
//wave sim vertex shader
attribute vec2 vPos;
varying vec2 uv;

void main() {
	uv = vPos.xy*.5+.5;
	gl_Position = vec4(vPos.xy,0,1);
}
</script>

<script type="text/plain" id="WaveSim.glsl">
//wave simulation(for death fx) pixel shader
varying vec2 uv;
uniform float ext;
uniform sampler2D sim;

vec4 Load(vec2 o) {
	vec4 v = texture2D(sim, uv+o);
	v.xz *= vec2(lessThan(v.yw,vec2(.5)))*-2.+1.;
	return v;
}

void main() {
	vec4 c = Load(vec2(0)),
		n = Load(vec2(0,ext)),
		e = Load(vec2(ext,0)),
		s = Load(vec2(0,-ext)),
		w = Load(vec2(-ext,0));

	c.z += (n.x+e.x+s.x+w.x)*.25-c.x;
	c.x += c.z;
	
	#define MIN_VAL 1./256.
	c.yw = vec2(c.x<0.?0.:1., c.z<0.?0.:1.);
	c.xz = abs(c.xz);
	if (c.x < MIN_VAL*6.) c.x = max(c.x-MIN_VAL, 0.);
	
	gl_FragColor = c;
}
</script>


<script type="text/plain" id="DeathEffectsVert.glsl">
//death effects vertex shader
attribute vec2 vPos;
varying vec2 uv;

uniform vec4 camera;//xy = camera position, zw = 1/(aspect*zoom)

void main() {
	vec2 wpos = vPos/camera.zw+camera.xy;
	uv = wpos/2048.;
	gl_Position = vec4(vPos.xy,0,1);
}
</script>

<script type="text/plain" id="DeathEffects.glsl">
//death effects pixel shader
varying vec2 uv;
uniform sampler2D sim;

void main() {
	vec2 state = texture2D(sim,uv).xy;
	if (state.x == 0.) {
		gl_FragColor = vec4(0);
		return;
	}
	float hue = (state.x*(state.y-.5)+.5)*40.+length(uv-.5)*300.,
		sat = abs(mod(state.x*13., 2.)-1.);
	vec3 waveCol = clamp(abs(mod(hue+vec3(0.,4.,2.),6.)-3.)-1., 0.,1.)*(1.-sat) + sat;
	gl_FragColor = vec4(waveCol*state.x, 0);
}
</script>



<script type="text/javascript">
