//Ethan Alexander Shulman 2025

var loadingPanel, mainMenuPanel, campaignPanel, customPanel, gamePanel, pausePanel, gameOverPanel,
aboutPanel, tutorialPanel, campaignLevelPanel, finalePanel, buyGamePanel,
quitBtn, pauseBtn, chooseColorBtn, musicVolume, deathEffectQuality,
gameFactionCounts, gameSelCount, tutorialTxt, tutorialDone,
customMapSize, customMapSeed, customFactions, customColors, customFactionCount = 0,
customAiInfo, customAiRows = [], customGameCode, finaleText,
campaignLvlBtns = [], campaignLevelTitle, campaignLevelInfo, campaignLevel = 0,
campaignResetTimer = 0, campaignResetCount = 0;

const DIFFICULTY_TAGS = ["<span style='color:#10ff10;'>Easy</span>",
"<span style='color:#ffff00;'>Medium</span>",
"<span style='color:#ff9010;'>Hard</span>",
"<span style='color:#ff1010;'>Insane</span>"];


function InitializeUI() {
	loadingPanel = DGE("LoadingPanel");
	
	
	//main menu panel
	mainMenuPanel = DGE("MainMenuPanel");
	
	BEC("TutorialBtn", function() {
		PlayBtnSfx();
		EV(mainMenuPanel, false);
		LaunchTutorial();
	});
	BEC("CampaignBtn", function() {
		//open campaign menu
		PlayBtnSfx();
		EV(mainMenuPanel, false);
		EV(campaignPanel, true);
	});
	BEC("CustomBtn", function() {
		//open custom menu
		PlayBtnSfx();
		UpdateAIColors();
		EV(mainMenuPanel, false);
		EV(customPanel, true);
	});
	BEC("QuitBtn", function() {
		SaveGame();
		if (NWJS_VERSION) {
			window["NWJSAppQuit"]();
		}
		if (WEBVIEW_VERSION) {
			window.open("https://quitapp.com", "_blank");
		}
	});
	if (!NWJS_VERSION && !WEBVIEW_VERSION) EV(DGE("QuitBtn"), false);
	
	BEC("QuickplayBtn", function() {
		//quickplay
		PlayBtnSfx();
		EV(mainMenuPanel, false);
		QuickPlay();
	});
	var aboutToggleFunc = function() {
		//about
		PlayBtnSfx();
		var aboutOn = (aboutPanel.style.display==="none");
		EV(aboutPanel, aboutOn);
		EV(mainMenuPanel, !aboutOn);
	};
	BEC("AboutBtn", aboutToggleFunc);
	
	BEC("MuteBtnsBtn", function(e) {
		btnsMuted = !btnsMuted;
		UpdateMuteBtn(e.target);
	});
	
	chooseColorBtn = DGE("ChooseColorBtn");
	chooseColorBtn.addEventListener("click", function() {
		//choose pixel color
		PlayBtnSfx();
		playerFactionId = (playerFactionId+1)%FACTION_COUNT;
		UpdatePlayerFaction();
	});
	
	
	//campaign panel
	campaignPanel = DGE("CampaignPanel");
	
	//back to main menu from campaign
	BEC("CampaignBackBtn", function() {
		PlayBtnSfx();
		EV(campaignPanel,false);
		EV(mainMenuPanel,true);
	});
	
	//create campaign level buttons
	var camLvls = DGE("CampaignLevels");
	for (var i = 0; i < CAMPAIGN_LEVELS.length; i++) {
		var ele = document.createElement("span");
		ele.className += "Btn CBtn";
		ele.style.color = "#606060";
		ele.textContent = (i+1)+"";
		ele.campaignLvlId = i;
		ele.addEventListener("click", SelectCampaignLevel);
		campaignLvlBtns.push(ele);
		camLvls.appendChild(ele);
	}
	
	var saveInput = DGE("CampaignSave"),
			saveTxt = DGE("CampaignSaveTxt");
	//export save
	BEC("CampaignExport", function() {
		saveInput.value = ExportCampaignProgress();
		saveInput.select();
		saveTxt.innerText = "Progress code copied to clipboard.";
		document.execCommand("copy");
	});

	//import save
	BEC("CampaignImport", function() {
		saveTxt.innerText = (ImportCampaignProgress(saveInput.value) ? "Imported campaign progress." : "Error invalid progress code.");
	});
	
	//reset campaign
	BEC("CampaignReset", function() {
		var timer = currentTime-campaignResetTimer;
		if (timer > 10000) {
			saveTxt.innerText = "Are you completely sure you want to reset your campaign progress?"+
				" Press Reset 3 times in quick succession to confirm.";
			campaignResetCount = 0;
			
		} else if (timer < 500) {
			campaignResetCount++;
			if (++campaignResetCount >= 3) {
				var nTimes = Math.min(100,campaignTimes.length+1);
				for (var i = 1; i < nTimes; i++) UnlockCampaignLevel(i, false);
				campaignTimes.length = 0;	
				SaveGame();
			
				saveTxt.innerText = "Campaign progress reset.";
				campaignResetCount = -1e4;
			}
			
		} else {
			if (campaignResetCount > 0) campaignResetCount = 0;
		}
		campaignResetTimer = currentTime;
	});
	
	
	//campaign level panel
	campaignLevelPanel = DGE("CampaignLevelPanel");
	campaignLevelTitle = DGE("LevelTitle");
	campaignLevelInfo = DGE("LevelInfo");
	
	BEC("LevelBackBtn", function() {
		PlayBtnSfx();
		EV(campaignLevelPanel,false);
		EV(campaignPanel,true);
	});
	
	BEC("PlayLevelBtn", function() {
		//launch level
		PlayBtnSfx();
		EV(campaignLevelPanel,false);
		playingCampaignLevel = campaignLevel;
		var ld = CAMPAIGN_LEVELS[campaignLevel];
		LaunchGame(ld[0], ld[1], ld[3], ld[2]);
	});
	
	
	//about panel
	aboutPanel = DGE("AboutPanel");
	BEC("AboutBackBtn", aboutToggleFunc);
	
	
	//custom game panel
	customPanel = DGE("CustomPanel");

	customMapSize = DGE("CustomMapSize");
	PopulateSelect(customMapSize, MAP_SIZES);

	customMapSeed = DGE("CustomMapSeed");
	customMapSeed.value = GenInitSeed(8);

	customFactions = DGE("CustomFactions");
	PopulateSelect(customFactions, ["2","3","4","5","6"]);
	customFactions.addEventListener("change", ResizeCustomAISelect);

	customColors = DGE("CustomColors");
	PopulateSelect(customColors, ["#1","#2","#3","#4","#5"]);
	customColors.addEventListener("change", UpdateAIColors);

	customAiInfo = DGE("CustomAIInfo");
	customGameCode = DGE("CustomGameCode");

	ResizeCustomAISelect();

	BEC("CustomCancel", function() {
		//close custom panel
		PlayBtnSfx();
		EV(customPanel, false);
		EV(mainMenuPanel, true);
	});
	
	BEC("CustomExport", function() {
		//export custom game as shareable code
		customGameCode.value = ExportGameCode();
		customAiInfo.innerText = "Game Code copied to clipboard, import the code on another machine.";
		customGameCode.select();
		document.execCommand("copy");
	});
	
	BEC("CustomImport", function() {
		customAiInfo.innerText = (ImportGameCode() ? "Imported Game Code." : "Invalid Game Code.");
	});
	
	BEC("CustomRandomize", function() {
		//randomize custom game settings
		customMapSize.selectedIndex = Math.floor(Math.random()*3);
		customMapSeed.value = GenInitSeed(8);
		customColors.selectedIndex = Math.floor(Math.random()*5);
		for (var i = 0; i < customFactionCount; i++) {
			var ar = customAiRows[i];
			ar[1].selectedIndex = Math.floor(Math.random()*AI_INFO.length);
			ar[2].value = (0.1+Math.pow(Math.random(),2)*9.9).toFixed(1);
			ar[3].value = GenInitSeed(8);
		}
		UpdateAIInfo({"target": customAiRows[0][1]});
		UpdateAIColors();
	});
	
	BEC("CustomStart", function() {
		//launch custom game
		PlayBtnSfx();
		EV(customPanel, false);
		var ais = [];
		for (var i = 0; i < customFactionCount; i++) {
			var ar = customAiRows[i],
				delay = parseFloat(ar[2].value);
			if (isNaN(delay)) delay = 0;
			delay = Math.max(0.1,delay);
			ais.push([ar[1].selectedIndex, delay, SeedUint(ar[3].value)]);
		}
		LaunchGame(customMapSize.selectedIndex, SeedUint(customMapSeed.value), ais, customColors.selectedIndex);
	});

	UpdateAIInfo({"target":customAiRows[0][1]});

	
	//in game panel
	gamePanel = DGE("GamePanel");
	gameSelCount = DGE("GameSelCount");
	if (WEBVIEW_VERSION) gameSelCount.style.right = "1.5vw";
	
	gameFactionCounts = new Array(FACTION_COUNT);
	for (var i = 0; i < FACTION_COUNT; i++) {
		var se = document.createElement("span");
		se.className += "FactionCount";
		se.style.left = (WEBVIEW_VERSION ? "1.5vw" : "0.5vw");
		se.style.bottom = (0.5+i*2.5)+"vw";
		gamePanel.appendChild(se);
		gameFactionCounts[i] = se;
	}
	
	pauseBtn = DGE("GamePause");
	pauseBtn.addEventListener("click", function() {
		Pause();
	});
	
	//pause panel
	pausePanel = DGE("PausePanel");
	musicVolume = DGE("PauseVolume");
	deathEffectQuality = DGE("PauseEffect");
	var blockInputPropagation = function(e) {
		e.stopPropagation();
	};
	musicVolume.addEventListener("mousedown", blockInputPropagation);
	//musicVolume.addEventListener("touchdown", blockInputPropagation);
	deathEffectQuality.addEventListener("mousedown", blockInputPropagation);
	//deathEffectQuality.addEventListener("touchdown", blockInputPropagation);
	
	quitBtn = DGE("PauseQuit");
	quitBtn.addEventListener("click", function() {
		UpdateDeathFxQuality(parseInt(deathEffectQuality.value));
		UpdateVolume();
		SaveGame();
		PlayBtnSfx();
		QuitGame();
	});
	
	
	//game over panel
	gameOverPanel = DGE("GameOverPanel");
	
	BEC("GameOverReturn", function() {
		cameraX = 1024;
		cameraY = 1024;
		cameraZoom = 256;
		ClearMap();
		ClearPixels();
		musicAudio.pause();
		PlayBtnSfx();
		
		appStatus = APPSTATUS_MAINMENU;
		EV(gameOverPanel,false);
		EV(mainMenuPanel,true);
	});
	
	//finale
	finalePanel = DGE("FinalePanel");
	finaleText = DGE("FinaleText");
	
	//tutorial panel
	tutorialPanel = DGE("TutorialPanel");
	tutorialTxt = DGE("TutorialTxt");
	tutorialDone = DGE("TutorialDone");
	
	BEC("TutorialQuit", function() {
		PlayBtnSfx();
		inTutorial = false;
		EV(tutorialPanel, false);
		QuitGame();
	});
}



/*Function: FatalCrash
Display fullscreen crash message.

Parameters:
*string* msg - HTML message.*/
function FatalCrash(msg) {
	document.body.innerHTML = "<div class='FatalCrash'>"+msg+"</div>";
}


//shortcut to get element by id
var DGE = document.getElementById.bind(document);

//shortcut for getting element by id and adding click event listener
function BEC(id,cb) {
	document.getElementById(id).addEventListener("click",cb);
}

//set element visible
function EV(ele,vis,inlineBlock,inline) {
	ele.style.display = vis?(inlineBlock?(inline?"inline":"inline-block"):"block"):"none";
}

//populate select element with options
function PopulateSelect(selEle,opt) {
	for (var i = 0; i < opt.length; i++) {
		var e = document.createElement("option");
		e.innerText = opt[i];
		selEle.appendChild(e);
	}
}


//rgb to css color string
function RGBToCSS(r,g,b) {
	return "rgb("+Math.round(r*255)+","+Math.round(g*255)+","+Math.round(b*255)+")";
}


//generate initial seed string
function GenInitSeed(count) {
	var s = "";
	for (var i = 0; i < count; i++) {
		s += String.fromCharCode(Math.floor(32+Math.random()*95));
	}
	return s;
}

//convert seed string to uint32
function SeedUint(seed) {
	ubuf[0] = 0;
	for (var i = 0; i < seed.length; i++) {
		var bit = (i%4)*8,
			upshift = (Math.floor(i/4)%2)*2;
		ubuf[0] ^= ((((seed.charCodeAt(i)+224)<<upshift)&0xFF)<<bit);
	}
	return ubuf[0];
}


//uint to string
function UintToString(seed) {
	return String.fromCharCode(seed&0xFF) + String.fromCharCode((seed>>>8)&0xFF) +
		String.fromCharCode((seed>>>16)&0xFF) + String.fromCharCode((seed>>>24)&0xFF);
}

//string to uint
function StringToUint(str, ind) {
	return (str.charCodeAt(ind)&0xFF) | ((str.charCodeAt(ind+1)&0xFF)<<8) |
		((str.charCodeAt(ind+2)&0xFF)<<16) | ((str.charCodeAt(ind+3)&0xFF)<<24);
}

//uint seed to string
function SeedToString(seedStr) {
	return UintToString(SeedUint(seedStr));
}

//convert uint seed back to string
function UintSeedToString(seed) {
	var str = "";
	for (var i = 0; i < 4; i++) {
		var bit = i*8,
			low = (seed>>>bit)&0x3F;
		str += String.fromCharCode(32+low);
		seed ^= low<<bit;
	}
	if (!seed) return str;
	
	for (var i = 0; i < 4; i++) {
		var bit = i*8+6,
			hi = (seed>>>bit)&3;
		str += String.fromCharCode(32+(hi<<4));
	}
	return str;
}


//generate custom game code string
function ExportGameCode() {
	var code = String.fromCharCode(customMapSize.selectedIndex | (customFactionCount << 2) | (customColors.selectedIndex << 5));
	code += SeedToString(customMapSeed.value);
	for (var i = 0; i < customFactionCount; i++) {
		var ar = customAiRows[i];
		code += String.fromCharCode(ar[1].selectedIndex)+
				UintToString(Math.round(parseFloat(ar[2].value)*10))+
				SeedToString(ar[3].value);
	}
	return btoa(code);
}

//import custom game code
function ImportGameCode() {
	var code;
	try {
		code = atob(customGameCode.value);
	} catch(e) {
		return false;
	}
	if (code.length < 5) return false;
	
	var ch = code.charCodeAt(0),
		mapSize = ch&3, facCount = (ch>>>2)&7,
		colorInd = (ch>>>5)&7;
	if (mapSize > 2 || facCount > 5 || colorInd > 4 ||
		code.length < 5+facCount*9) return false;
	
	customMapSize.selectedIndex = mapSize;
	customFactions.selectedIndex = facCount-1;
	customColors.selectedIndex = colorInd;
	customMapSeed.value = UintSeedToString(StringToUint(code,1));	
	ResizeCustomAISelect();
	UpdateAIColors();

	var ci = 5;
	for (var i = 0; i < customFactionCount; i++) {
		var ar = customAiRows[i];
		ar[1].selectedIndex = Math.min(AI_INFO.length-1, code.charCodeAt(ci++));
		ar[2].value = (StringToUint(code,ci)/10).toFixed(1);
		ci += 4;
		ar[3].value = UintSeedToString(StringToUint(code,ci));
		ci += 4;
	}
	return true;
}


//export campaign progress
function ExportCampaignProgress() {
	var code = "";
	for (var i = 0; i < campaignTimes.length; i++) {
		code += UintToString(Math.round(campaignTimes[i]));
	}
	return btoa(code);
}

//import campaign progress
function ImportCampaignProgress(txt) {
	var code;
	try {
		code = atob(txt);
	} catch (e) {
		return false;
	}
	
	var nTimes = code.length/4;
	if (code.length%4 !== 0 || nTimes > 100) return false;
	
	var oldTimes = campaignTimes.length;
	campaignTimes.length = Math.max(oldTimes, nTimes);
	for (var i = 0; i < nTimes; i++) {
		ubuf[0] = StringToUint(code, i*4);
		campaignTimes[i] = Math.min((i<oldTimes?campaignTimes[i]:0xFFFFFFFF), ubuf[0]);
	}
	if (oldTimes < nTimes) {
		nTimes = Math.min(nTimes,99);
		for (var i = oldTimes; i < nTimes; i++) UnlockCampaignLevel(i+1, true);
	}
	SaveGame();
	return true;
}


//time(in seconds) as string
function TimeString(timeSec) {
	return (timeSec>59 ? Math.floor(timeSec/60)+"m " : "") + (timeSec%60)+"s";
}


//resize number of rows in custom game ai selection
function ResizeCustomAISelect() {
	var tbl = document.getElementById("CustomFactionTable"),
		aiCount = customFactions.selectedIndex+1;
	for (var i = customFactionCount; i < aiCount; i++) {
		var row = document.createElement("tr"),
			col = document.createElement("td"),
			aiCol = document.createElement("span"),
			aiSel = document.createElement("select"),
			aiDelay = document.createElement("input"),
			aiSeed = document.createElement("input");
		aiCol.className += "AIColor";
		col.appendChild(aiCol);
		row.appendChild(col);
			
		col = document.createElement("td");
		PopulateSelect(aiSel, AI_NAMES);
		aiSel.addEventListener("change", UpdateAIInfo);
		col.appendChild(aiSel);		
		row.appendChild(col);
		
		col = document.createElement("td");
		col.appendChild(document.createTextNode("AI Delay "));
		aiDelay.type = "text";
		aiDelay.size = 4;
		aiDelay.value = "1";
		col.appendChild(aiDelay);
		row.appendChild(col);
		tbl.appendChild(row);
		
		col = document.createElement("td");
		col.appendChild(document.createTextNode("AI Seed "));
		aiSeed.type = "text";
		aiSeed.size = 8;
		aiSeed.value = GenInitSeed(8);
		col.appendChild(aiSeed);
		row.appendChild(col);
		tbl.appendChild(row);
		
		customAiRows.push([aiCol,aiSel,aiDelay,aiSeed]);
	}
	
	for (var i = customFactionCount; i > aiCount; i--) {
		tbl.removeChild(tbl.lastChild);
	}
	customAiRows.length = aiCount;
	customFactionCount = aiCount;
	UpdateAIColors();
}

//update custom game ai color indicators
function UpdateAIColors() {
	var cind = customColors.selectedIndex;
	for (var i = 0; i < customFactionCount; i++) {
		var aiCol = customAiRows[i][0],
			aiId = (i+cind)%(FACTION_COUNT-1);
		aiCol.style.backgroundColor = FACTIONS[(playerFactionId+1+aiId)%FACTION_COUNT].cssColor;
	}
}


//unlock campaign level button
function UnlockCampaignLevel(lvl, unlock) {
	var btn = campaignLvlBtns[lvl];
	btn.style.color = (unlock ? "#ffffff" :  "#606060");
}

//select campaign level
function SelectCampaignLevel(e) {
	campaignLevel = e.target.campaignLvlId;
	if (campaignLevel > campaignTimes.length && !DEV_VERSION) {
		PlayBtnOffSfx();
		return;
	}
	PlayBtnSfx();
	
	campaignLevelTitle.textContent = "Level #"+(campaignLevel+1);
	
	var ld = CAMPAIGN_LEVELS[campaignLevel],
		aiTxt = "", ais = ld[3];
	for (var i = 0; i < ais.length; i++) {
		var aii = AI_INFO[ais[i][0]];
		if (i !== 0) aiTxt += ", ";
		aiTxt += aii[0]+"("+DIFFICULTY_TAGS[aii[2]]+")";
	}
	campaignLevelInfo.innerHTML = "Map Size: "+MAP_SIZES[ld[0]]+"</br>"+
								"Enemy AI: "+aiTxt+"</br>"+
								"Fastest Completion Time: "+
								(campaignTimes.length > campaignLevel ?
								TimeString(Math.round(campaignTimes[campaignLevel])) : "No time.");
	
	EV(campaignPanel, false);
	EV(campaignLevelPanel, true);
}


//update player faction
function UpdatePlayerFaction() {
	var playerCol = FACTIONS[playerFactionId].cssColor;
	chooseColorBtn.style.color = playerCol;
	gameSelCount.style.color = playerCol;
}

//update selected pixel count
function UpdateSelectCount() {
	var sl = selectedPixels.length;
	gameSelCount.textContent = (sl > 0 ? sl+" selected" : "");
}


//display ai info when selecting ai
function UpdateAIInfo(e) {
	var aii = AI_INFO[e.target.selectedIndex];
	customAiInfo.innerHTML = aii[0]+" "+aii[1]+"</br>Difficulty: "+DIFFICULTY_TAGS[aii[2]];
}


//toggle pause/unpaused
function Pause() {
	if (inTutorial) return;
	
	unpaused = !unpaused;
	pauseBtn.textContent = (unpaused ? "Pause" : "Unpause");
	if (unpaused) {
		//load changed settings
		UpdateDeathFxQuality(parseInt(deathEffectQuality.value));
		UpdateVolume();
		if (songDelay <= 0) musicAudio.play();
		
	} else {
		if (songDelay <= 0) musicAudio.pause();
	}
	EV(pausePanel, !unpaused);
}


//update mute btn text
function UpdateMuteBtn(btn) {
	btn.innerHTML = (btnsMuted ? "Unmute</br>Buttons" : "Mute</br>Buttons");
}
